"""
Copied from https://github.com/tencent-ailab/grover/blob/0421d97a5e1bd1b59d1923e3afd556afbe4ff782/grover/model/layers.py
"""
from typing import List, Dict
try:
    import torch
    import torch.nn as nn
except ModuleNotFoundError:
    raise ImportError('These classes require PyTorch to be installed.')
import numpy as np
from scipy import stats
from deepchem.models.torch_models.readout import GroverReadout
from deepchem.models.torch_models.layers import SublayerConnection, PositionwiseFeedForward


class GroverEmbedding(nn.Module):
    """GroverEmbedding layer.

    This layer is a simple wrapper over GroverTransEncoder layer for retrieving the embeddings from the GroverTransEncoder layer.

    Parameters
    ----------
    edge_fdim: int
        the dimension of additional feature for edge/bond.
    node_fdim: int
        the dimension of additional feature for node/atom.
    depth: int
        Dynamic message passing depth for use in MPNEncoder
    undirected: bool
        The message passing is undirected or not
    num_mt_block: int
        the number of message passing blocks.
    num_head: int
        the number of attention heads.
    """

    def __init__(self,
                 node_fdim,
                 edge_fdim,
                 hidden_size=128,
                 depth=1,
                 undirected=False,
                 dropout=0.2,
                 activation='relu',
                 num_mt_block=1,
                 num_heads=4,
                 bias=False,
                 res_connection=False):
        super(GroverEmbedding, self).__init__()
        self.encoders = GroverTransEncoder(hidden_size=hidden_size,
                                           edge_fdim=edge_fdim,
                                           node_fdim=node_fdim,
                                           depth=depth,
                                           undirected=undirected,
                                           dropout=dropout,
                                           activation=activation,
                                           num_mt_block=num_mt_block,
                                           num_heads=num_heads,
                                           bias=bias,
                                           res_connection=res_connection)

    def forward(self, graph_batch: List[torch.Tensor]):
        """Forward function

        Parameters
        ----------
        graph_batch: List[torch.Tensor]
            A list containing f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, a2a

        Returns
        -------
        embedding: Dict[str, torch.Tensor]
            Returns a dictionary of embeddings. The embeddings are:
         - atom_from_atom: node messages aggregated from node hidden states
         - bond_from_atom: bond messages aggregated from bond hidden states
         - atom_from_bond: node message aggregated from bond hidden states
         - bond_from_bond: bond messages aggregated from bond hidden states.
        """
        # TODO Explain in detail what the four outcompes are
        output = self.encoders(graph_batch)
        return {
            "atom_from_atom": output[0][0],
            "bond_from_atom": output[0][1],
            "atom_from_bond": output[1][0],
            "bond_from_bond": output[1][1]
        }


class GroverBondVocabPredictor(nn.Module):
    """Layer for learning contextual information for bonds.

    The layer is used in Grover architecture to learn contextual information of a bond by predicting
    the context of a bond from the bond embedding in a multi-class classification setting.
    The contextual information of a bond are encoded as strings (ex: '(DOUBLE-STEREONONE-NONE)_C-(SINGLE-STEREONONE-NONE)2').

    Example
    -------
    >>> from deepchem.models.torch_models.grover_layers import GroverBondVocabPredictor
    >>> num_bonds = 20
    >>> in_features, vocab_size = 16, 10
    >>> layer = GroverBondVocabPredictor(vocab_size, in_features)
    >>> embedding = torch.randn(num_bonds * 2, in_features)
    >>> result = layer(embedding)
    >>> result.shape
    torch.Size([20, 10])

    Reference
    ---------
    .. Rong, Yu, et al. "Self-supervised graph transformer on large-scale molecular data." Advances in Neural Information Processing Systems 33 (2020): 12559-12571.
    """

    def __init__(self, vocab_size: int, in_features: int = 128):
        """Initializes GroverBondVocabPredictor

        Parameters
        ----------
        vocab_size: int
            Size of vocabulary, used for number of classes in prediction.
        in_features: int, default: 128
            Input feature size of bond embeddings.
        """
        super(GroverBondVocabPredictor, self).__init__()
        self.linear = nn.Linear(in_features, vocab_size)
        self.linear_rev = nn.Linear(in_features, vocab_size)
        self.logsoftmax = nn.LogSoftmax(dim=1)

    def forward(self, embeddings):
        """
        Parameters
        ----------
        embeddings: torch.Tensor
            bond embeddings of shape (num_bond, in_features)

        Returns
        -------
        logits: torch.Tensor
            the prediction for each bond, (num_bond, vocab_size)
        """
        nm_bonds = embeddings.shape[0]
        # The bond and rev bond have even and odd ids respectively.
        ids1 = list(range(0, nm_bonds, 2))
        ids2 = list(range(1, nm_bonds, 2))
        logits = self.linear(embeddings[ids1]) + self.linear_rev(
            embeddings[ids2])
        return self.logsoftmax(logits)


class GroverAtomVocabPredictor(nn.Module):
    """Grover Atom Vocabulary Prediction Module.

    The GroverAtomVocabPredictor module is used for predicting atom-vocabulary
    for the self-supervision task in Grover architecture. In the self-supervision tasks,
    one task is to learn contextual-information of nodes (atoms).
    Contextual information are encoded as strings, like `C_N-DOUBLE1_O-SINGLE1`.
    The module accepts an atom encoding and learns to predict the contextual information
    of the atom as a multi-class classification problem.

    Example
    -------
    >>> from deepchem.models.torch_models.grover_layers import GroverAtomVocabPredictor
    >>> num_atoms, in_features, vocab_size = 30, 16, 10
    >>> layer = GroverAtomVocabPredictor(vocab_size, in_features)
    >>> embedding = torch.randn(num_atoms, in_features)
    >>> result = layer(embedding)
    >>> result.shape
    torch.Size([30, 10])

    Reference
    ---------
    .. Rong, Yu, et al. "Self-supervised graph transformer on large-scale molecular data." Advances in Neural Information Processing Systems 33 (2020): 12559-12571.
    """

    def __init__(self, vocab_size: int, in_features: int = 128):
        """Initializing Grover Atom Vocabulary Predictor

        Parameters
        ----------
        vocab_size: int
            size of vocabulary (vocabulary here is the total number of different possible contexts)
        in_features: int
            feature size of atom embeddings.
        """
        super(GroverAtomVocabPredictor, self).__init__()
        self.linear = nn.Linear(in_features, vocab_size)
        self.logsoftmax = nn.LogSoftmax(dim=1)

    def forward(self, embeddings):
        """
        Parameters
        ----------
        embeddings: torch.Tensor
            the atom embeddings of shape (vocab_size, in_features)

        Returns
        -------
        logits: torch.Tensor
            the prediction for each atom of shape (num_bond, vocab_size)
        """
        return self.logsoftmax(self.linear(embeddings))


class GroverFunctionalGroupPredictor(nn.Module):
    """The functional group prediction task for self-supervised learning.

    Molecules have functional groups in them. This module is used for predicting
    the functional group and the problem is formulated as an multi-label classification problem.

    Parameters
    ----------
    functional_group_size: int,
        size of functional group
    in_features: int,
        hidden_layer size, default 128

    Example
    -------
    >>> from deepchem.models.torch_models.grover_layers import GroverFunctionalGroupPredictor
    >>> in_features, functional_group_size = 8, 20
    >>> num_atoms, num_bonds = 10, 20
    >>> predictor = GroverFunctionalGroupPredictor(functional_group_size=20, in_features=8)
    >>> atom_scope, bond_scope = [(0, 3), (3, 3), (6, 4)], [(0, 5), (5, 4), (9, 11)]
    >>> embeddings = {}
    >>> embeddings['bond_from_atom'] = torch.randn(num_bonds, in_features)
    >>> embeddings['bond_from_bond'] = torch.randn(num_bonds, in_features)
    >>> embeddings['atom_from_atom'] = torch.randn(num_atoms, in_features)
    >>> embeddings['atom_from_bond'] = torch.randn(num_atoms, in_features)
    >>> result = predictor(embeddings, atom_scope, bond_scope)

    Reference
    ---------
    .. Rong, Yu, et al. "Self-supervised graph transformer on large-scale molecular data." Advances in Neural Information Processing Systems 33 (2020): 12559-12571.

    """

    def __init__(self, functional_group_size: int, in_features=128):
        super(GroverFunctionalGroupPredictor, self).__init__()

        self.readout = GroverReadout(rtype="mean", in_features=in_features)
        self.linear_atom_from_atom = nn.Linear(in_features,
                                               functional_group_size)
        self.linear_atom_from_bond = nn.Linear(in_features,
                                               functional_group_size)
        self.linear_bond_from_atom = nn.Linear(in_features,
                                               functional_group_size)
        self.linear_bond_from_bond = nn.Linear(in_features,
                                               functional_group_size)

    def forward(self, embeddings: Dict, atom_scope: List, bond_scope: List):
        """
        The forward function for the GroverFunctionalGroupPredictor (semantic motif prediction) layer.
        It takes atom/bond embeddings produced from node and bond hidden states from GroverEmbedding module
        and the atom, bond scopes and produces prediction logits for different each embedding.
        The scopes are used to differentiate atoms/bonds belonging to a molecule in a batched molecular graph.

        Parameters
        ----------
        embedding: Dict
            The input embeddings organized as an dictionary. The input embeddings are output of GroverEmbedding layer.
        atom_scope: List
            The scope for atoms.
        bond_scope: List
            The scope for bonds

        Returns
        -------
        preds: Dict
            A dictionary containing the predicted logits of functional group from four different types of input embeddings. The key and their corresponding predictions
        are described below.
         - atom_from_atom - prediction logits from atom embeddings generated via node hidden states
         - atom_from_bond - prediction logits from atom embeddings generated via bond hidden states
         - bond_from_atom - prediction logits from bond embeddings generated via node hidden states
         - bond_from_bond - prediction logits from bond embeddings generated via bond hidden states
        """
        preds_bond_from_atom = self.linear_bond_from_atom(
            self.readout(embeddings["bond_from_atom"], bond_scope))
        preds_bond_from_bond = self.linear_bond_from_bond(
            self.readout(embeddings["bond_from_bond"], bond_scope))

        preds_atom_from_atom = self.linear_atom_from_atom(
            self.readout(embeddings["atom_from_atom"], atom_scope))
        preds_atom_from_bond = self.linear_atom_from_bond(
            self.readout(embeddings["atom_from_bond"], atom_scope))

        return {
            "atom_from_atom": preds_atom_from_atom,
            "atom_from_bond": preds_atom_from_bond,
            "bond_from_atom": preds_bond_from_atom,
            "bond_from_bond": preds_bond_from_bond
        }


def _index_select_nd(source: torch.Tensor, index: torch.Tensor) -> torch.Tensor:
    """
    Selects the message features from source corresponding to the atom or bond indices in index.

    Parameters
    ----------
    source: torch.Tensor
        A tensor of shape (num_bonds, hidden_size) containing message features.
    index: torch.Tensor
        A tensor of shape (num_atoms/num_bonds, max_num_bonds) containing the atom or bond indices to select from source.

    Returns
    ----------
    message_features: torch.Tensor
        A tensor of shape (num_atoms/num_bonds, max_num_bonds, hidden_size) containing the message features corresponding to the atoms/bonds specified in index.
    """
    index_size = index.size()  # (num_atoms/num_bonds, max_num_bonds)
    suffix_dim = source.size()[1:]  # (hidden_size,)
    final_size = index_size + suffix_dim  # (num_atoms/num_bonds, max_num_bonds, hidden_size)

    target = source.index_select(dim=0, index=index.view(
        -1))  # (num_atoms/num_bonds * max_num_bonds, hidden_size)
    target = target.view(
        final_size)  # (num_atoms/num_bonds, max_num_bonds, hidden_size)

    return target


def _select_neighbor_and_aggregate(feature, index):
    """The basic operation in message passing.

    Caution: the index_selec_ND would cause the reproducibility issue when performing the training on CUDA.

    The operation is like map and reduce. `index_select_nd` maps message features to bonds/atoms and
    this method aggregates the results by using `sum` as pooling operation. We can also add a configuration
    here to use `mean` as pooling operation but it is left to future implementation.

    References
    ----------
    See: https://pytorch.org/docs/stable/notes/randomness.html

    Parameters
    ----------
    feature: np.array
        The candidate feature for aggregate. (n_nodes, hidden)
    index: np.array
        The selected index (neighbor indexes).

    Returns
    -------
    None
    """
    neighbor = _index_select_nd(feature, index)
    return neighbor.sum(dim=1)


class GroverMPNEncoder(nn.Module):
    """Performs Message Passing to generate encodings for the molecule.

    Parameters
    ----------
    atom_messages: bool
        True if encoding atom-messages else False.
    init_message_dim: int
        Dimension of embedding message.
    attach_feats: bool
        Set to `True` if additional features are passed along with node/edge embeddings.
    attached_feat_fdim: int
        Dimension of additional features when `attach_feats` is `True`
    undirected: bool
        If set to `True`, the graph is considered as an undirected graph.
    depth: int
        number of hops in a message passing iteration
    dynamic_depth: str, default: none
        If set to `uniform` for randomly sampling dynamic depth from an uniform distribution else if set to `truncnorm`, dynamic depth is sampled from a truncated normal distribution.
    input_layer: str
        If set to `fc`, adds an initial feed-forward layer. If set to `none`, does not add an initial feed forward layer.
    """

    # FIXME This layer is similar to DMPNNEncoderLayer and they
    # must be unified.
    def __init__(self,
                 atom_messages: bool,
                 init_message_dim: int,
                 hidden_size: int,
                 depth: int,
                 undirected: bool,
                 attach_feats: bool,
                 attached_feat_fdim: int = 0,
                 bias: bool = True,
                 dropout: float = 0.2,
                 activation: str = 'relu',
                 input_layer: str = 'fc',
                 dynamic_depth: str = 'none'):
        super(GroverMPNEncoder, self).__init__()
        if input_layer == 'none':
            assert init_message_dim == hidden_size
        assert dynamic_depth in [
            'uniform', 'truncnorm', 'none'
        ], 'dynamic depth should be one of uniform, truncnorm, none'
        self.init_message_dim = init_message_dim
        self.depth = depth
        self.input_layer = input_layer
        self.layers_per_message = 1
        self.undirected = undirected
        self.atom_messages = atom_messages
        self.attached_feat = attach_feats

        assert dynamic_depth in [
            'none', 'truncnorm', 'uniform'
        ], 'If dynamic depth, it should be truncnorm or uniform'
        self.dynamic_depth = dynamic_depth

        self.dropout_layer = nn.Dropout(p=dropout)
        if activation == 'relu':
            self.act_func = nn.ReLU()
        else:
            raise ValueError('Only ReLU activation function is supported')

        if self.input_layer == "fc":
            self.W_i = nn.Linear(self.init_message_dim, hidden_size, bias=bias)
            w_h_input_size = hidden_size
        elif input_layer == 'none':
            w_h_input_size = self.init_message_dim

        if self.attached_feat:
            w_h_input_size += attached_feat_fdim

        # Shared weight matrix across depths (default)
        self.W_h = nn.Linear(w_h_input_size, hidden_size, bias=bias)

    def forward(self,
                init_messages,
                init_attached_features,
                a2nei,
                a2attached,
                b2a=None,
                b2revb=None,
                adjs=None) -> torch.FloatTensor:
        if self.input_layer == 'fc':
            message = self.act_func(self.W_i(init_messages))
        elif self.input_layer == 'none':
            message = init_messages

        attached_feats = init_attached_features

        if self.training and self.dynamic_depth != 'none':
            if self.dynamic_depth == 'uniform':
                ndepth = abs(
                    int(np.random.uniform(self.depth - 3, self.depth + 3)))
            elif self.dynamic_depth == 'truncnorm':
                mu, sigma = self.depth, 1
                lower, upper = mu - 3 * sigma, mu + 3 * sigma
                X = stats.truncnorm((lower - mu) / sigma, (upper - mu) / sigma,
                                    loc=mu,
                                    scale=sigma)
                ndepth = int(X.rvs(1)[0])
        else:
            ndepth = self.depth

        for _ in range(ndepth - 1):
            if self.undirected:
                message = (message + message[b2revb]) / 2

            nei_message = _select_neighbor_and_aggregate(message, a2nei)
            a_message = nei_message
            if self.attached_feat:
                attached_nei_feats = _select_neighbor_and_aggregate(
                    attached_feats, a2attached)
                a_message = torch.cat((nei_message, attached_nei_feats), dim=1)

            if not self.atom_messages:
                rev_message = message[b2revb]
                if self.attached_feat:
                    atom_rev_message = attached_feats[b2a[b2revb]]
                    rev_message = torch.cat((rev_message, atom_rev_message),
                                            dim=1)
                # Except reverse bond its-self(w) ! \sum_{k\in N(u) \ w}
                message = a_message[b2a] - rev_message  # num_bonds x hidden
            else:
                message = a_message

            # FIXME When input_layer is none, for the first iteration of message passing, we should ideally
            # be using different weight matrix since message will be of shape num_bonds x f_bonds_dim
            # in the first iteration and in the subsequent iterations, it will be num_bonds x hidden_size
            message = self.W_h(message)

            message = self.dropout_layer(self.act_func(message))

        return message


class GroverAttentionHead(nn.Module):
    """Generates attention head using GroverMPNEncoder for generating query, key and value

    Parameters
    ----------
    hidden_size: int
        Dimension of hidden layer
    undirected: bool
        If set to `True`, the graph is considered as an undirected graph.
    depth: int
        number of hops in a message passing iteration
    atom_messages: bool
        True if encoding atom-messages else False.
    """

    def __init__(self,
                 hidden_size: int = 128,
                 bias: bool = True,
                 depth: int = 1,
                 dropout: float = 0.0,
                 undirected: bool = False,
                 atom_messages: bool = False):
        super(GroverAttentionHead, self).__init__()
        self.atom_messages = atom_messages

        # FIXME We assume that we are using a hidden layer to transform the initial atom message
        # and bond messages to hidden dimension size.
        self.mpn_q = GroverMPNEncoder(atom_messages=atom_messages,
                                      init_message_dim=hidden_size,
                                      attached_feat_fdim=hidden_size,
                                      hidden_size=hidden_size,
                                      bias=bias,
                                      depth=depth,
                                      dropout=dropout,
                                      undirected=undirected,
                                      attach_feats=False,
                                      input_layer='none',
                                      dynamic_depth='truncnorm')

        self.mpn_k = GroverMPNEncoder(atom_messages=atom_messages,
                                      init_message_dim=hidden_size,
                                      attached_feat_fdim=hidden_size,
                                      hidden_size=hidden_size,
                                      bias=bias,
                                      depth=depth,
                                      dropout=dropout,
                                      undirected=undirected,
                                      attach_feats=False,
                                      input_layer='none',
                                      dynamic_depth='truncnorm')

        self.mpn_v = GroverMPNEncoder(atom_messages=atom_messages,
                                      init_message_dim=hidden_size,
                                      attached_feat_fdim=hidden_size,
                                      hidden_size=hidden_size,
                                      bias=bias,
                                      depth=depth,
                                      dropout=dropout,
                                      undirected=undirected,
                                      attach_feats=False,
                                      input_layer='none',
                                      dynamic_depth='truncnorm')

    def forward(self, f_atoms, f_bonds, a2b, a2a, b2a, b2revb):
        if self.atom_messages:
            init_messages = f_atoms
            init_attached_features = f_bonds
            a2nei = a2a
            a2attached = a2b
            b2a = b2a
            b2revb = b2revb
        else:
            # self.atom_messages is False
            init_messages = f_bonds
            init_attached_features = f_atoms
            a2nei = a2b
            a2attached = a2a
            b2a = b2a
            b2revb = b2revb

        q = self.mpn_q(init_messages=init_messages,
                       init_attached_features=init_attached_features,
                       a2nei=a2nei,
                       a2attached=a2attached,
                       b2a=b2a,
                       b2revb=b2revb)
        k = self.mpn_k(init_messages=init_messages,
                       init_attached_features=init_attached_features,
                       a2nei=a2nei,
                       a2attached=a2attached,
                       b2a=b2a,
                       b2revb=b2revb)
        v = self.mpn_v(init_messages=init_messages,
                       init_attached_features=init_attached_features,
                       a2nei=a2nei,
                       a2attached=a2attached,
                       b2a=b2a,
                       b2revb=b2revb)
        return q, k, v


class GroverMTBlock(nn.Module):
    """Message passing combined with transformer architecture

    The layer combines message passing performed using GroverMPNEncoder and uses it
    to generate query, key and value for multi-headed Attention block.

    Parameters
    ----------
    atom_messages: bool
        True if encoding atom-messages else False.
    input_dim: int
        Dimension of input features
    num_heads: int
        Number of attention heads
    depth: int
        Number of hops in a message passing iteration
    undirected: bool
        If set to `True`, the graph is considered as an undirected graph.
    """

    def __init__(self,
                 atom_messages: bool,
                 input_dim: int,
                 num_heads: int,
                 depth: int,
                 undirected: bool = False,
                 hidden_size: int = 128,
                 dropout: float = 0.0,
                 bias: bool = True,
                 res_connection: bool = True,
                 activation: str = 'relu'):
        super(GroverMTBlock, self).__init__()
        self.hidden_size = hidden_size
        self.atom_messages = atom_messages
        self.res_connection = res_connection
        self.num_heads = num_heads
        if activation == 'relu':
            self.act_func = nn.ReLU()
        else:
            raise ValueError('Only relu activation is supported')
        self.dropout_layer = nn.Dropout(p=dropout)

        # Note: Elementwise affine has to be consistent with the pre-training phase
        self.layernorm = nn.LayerNorm(hidden_size, elementwise_affine=True)
        self.attn = torch.nn.MultiheadAttention(embed_dim=hidden_size,
                                                num_heads=num_heads,
                                                dropout=dropout,
                                                bias=bias,
                                                batch_first=True)

        self.W_i = nn.Linear(input_dim, hidden_size, bias)
        self.W_o = nn.Linear(hidden_size * num_heads, hidden_size, bias)
        self.sublayer = SublayerConnection(hidden_size, dropout)
        self.attention_heads = nn.ModuleList()
        for i in range(num_heads):
            self.attention_heads.append(
                GroverAttentionHead(hidden_size=hidden_size,
                                    bias=bias,
                                    dropout=dropout,
                                    depth=depth,
                                    atom_messages=atom_messages,
                                    undirected=undirected))

    def forward(self, batch):
        f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, a2a = batch
        if self.atom_messages:
            if f_atoms.shape[1] != self.hidden_size:
                f_atoms = self.W_i(f_atoms)
                f_atoms = self.dropout_layer(
                    self.layernorm(self.act_func(f_atoms)))

        else:
            if f_bonds.shape[1] != self.hidden_size:
                f_bonds = self.W_i(f_bonds)
                f_bonds = self.dropout_layer(
                    self.layernorm(self.act_func(f_bonds)))

        queries, keys, values = [], [], []
        for head in self.attention_heads:
            q, k, v = head(f_atoms,
                           f_bonds,
                           a2b=a2b,
                           b2a=b2a,
                           b2revb=b2revb,
                           a2a=a2a)
            queries.append(q.unsqueeze(1))
            keys.append(k.unsqueeze(1))
            values.append(v.unsqueeze(1))
        queries = torch.cat(queries, dim=1)
        keys = torch.cat(keys, dim=1)
        values = torch.cat(values, dim=1)

        # multi-headed attention
        x_out, _ = self.attn(queries, keys, values)
        x_out = x_out.reshape(x_out.shape[0], -1)
        x_out = self.W_o(x_out)

        # support no residual connection in MTBlock.
        if self.res_connection:
            if self.atom_messages:
                f_atoms = self.sublayer(f_atoms, x_out)
            else:
                f_bonds = self.sublayer(f_bonds, x_out)
        batch = f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, a2a
        return batch


class GroverTransEncoder(nn.Module):
    """GroverTransEncoder for encoding a molecular graph

    The GroverTransEncoder layer is used for encoding a molecular graph.
    The layer returns 4 outputs. They are atom messages aggregated from atom hidden states,
    atom messages aggregated from bond hidden states, bond messages aggregated from atom hidden
    states, bond messages aggregated from bond hidden states.

    Parameters
    ----------
    hidden_size: int
        the hidden size of the model.
    edge_fdim: int
        the dimension of additional feature for edge/bond.
    node_fdim: int
        the dimension of additional feature for node/atom.
    depth: int
        Dynamic message passing depth for use in MPNEncoder
    undirected: bool
        The message passing is undirected or not
    dropout: float
        the dropout ratio
    activation: str
        the activation function
    num_mt_block: int
        the number of mt block.
    num_head: int
        the number of attention AttentionHead.
    bias: bool
        enable bias term in all linear layers.
    res_connection: bool
        enables the skip-connection in MTBlock.
    """

    def __init__(self,
                 node_fdim: int,
                 edge_fdim: int,
                 depth: int = 3,
                 undirected: bool = False,
                 num_mt_block: int = 2,
                 num_heads: int = 2,
                 hidden_size: int = 64,
                 dropout: float = 0.2,
                 res_connection: bool = True,
                 bias: bool = True,
                 activation: str = 'relu'):
        super(GroverTransEncoder, self).__init__()

        self.hidden_size = hidden_size
        self.edge_fdim = edge_fdim
        self.node_fdim = node_fdim

        self.edge_blocks = nn.ModuleList()
        self.node_blocks = nn.ModuleList()

        for i in range(num_mt_block):
            if i == 0:
                node_input_fdim, edge_input_fdim = node_fdim, edge_fdim
            else:
                node_input_fdim, edge_input_fdim = hidden_size, hidden_size

            self.edge_blocks.append(
                GroverMTBlock(num_heads=num_heads,
                              input_dim=edge_input_fdim,
                              hidden_size=hidden_size,
                              activation=activation,
                              dropout=dropout,
                              bias=bias,
                              atom_messages=False,
                              res_connection=res_connection,
                              depth=depth,
                              undirected=undirected))
            self.node_blocks.append(
                GroverMTBlock(num_heads=num_heads,
                              input_dim=node_input_fdim,
                              hidden_size=hidden_size,
                              activation=activation,
                              dropout=dropout,
                              bias=bias,
                              atom_messages=True,
                              res_connection=res_connection,
                              depth=depth,
                              undirected=undirected))

        self.ffn_atom_from_atom = PositionwiseFeedForward(
            d_input=self.hidden_size + node_fdim,
            d_hidden=self.hidden_size * 4,
            d_output=self.hidden_size,
            n_layers=2,
            activation=activation,
            dropout_p=dropout)

        self.ffn_atom_from_bond = PositionwiseFeedForward(
            d_input=self.hidden_size + node_fdim,
            d_hidden=self.hidden_size * 4,
            d_output=self.hidden_size,
            n_layers=2,
            activation=activation,
            dropout_p=dropout)

        self.ffn_bond_from_atom = PositionwiseFeedForward(
            d_input=self.hidden_size + edge_fdim,
            d_hidden=self.hidden_size * 4,
            d_output=self.hidden_size,
            n_layers=2,
            activation=activation,
            dropout_p=dropout)

        self.ffn_bond_from_bond = PositionwiseFeedForward(
            d_input=self.hidden_size + edge_fdim,
            d_hidden=self.hidden_size * 4,
            d_output=self.hidden_size,
            n_layers=2,
            activation=activation,
            dropout_p=dropout)

        self.atom_from_atom_sublayer = SublayerConnection(size=self.hidden_size,
                                                          dropout_p=dropout)
        self.atom_from_bond_sublayer = SublayerConnection(size=self.hidden_size,
                                                          dropout_p=dropout)
        self.bond_from_atom_sublayer = SublayerConnection(size=self.hidden_size,
                                                          dropout_p=dropout)
        self.bond_from_bond_sublayer = SublayerConnection(size=self.hidden_size,
                                                          dropout_p=dropout)

        if activation == 'relu':
            self.act_func_node = nn.ReLU()
            self.act_func_edge = nn.ReLU()
        else:
            raise ValueError('Only relu activation is supported')

        self.dropout_layer = nn.Dropout(p=dropout)

    def _pointwise_feed_forward_to_atom_embedding(self, emb_output, atom_feat,
                                                  index, ffn_layer):
        aggr_output = _select_neighbor_and_aggregate(emb_output, index)
        aggr_outputx = torch.cat([atom_feat, aggr_output], dim=1)
        return ffn_layer(aggr_outputx), aggr_outputx

    def _pointwise_feed_forward_to_bond_embedding(self, emb_output, bond_feat,
                                                  a2nei, b2revb, ffn_layer):
        aggr_output = _select_neighbor_and_aggregate(emb_output, a2nei)
        aggr_output = self._remove_rev_bond_message(emb_output, aggr_output,
                                                    b2revb)
        aggr_outputx = torch.cat([bond_feat, aggr_output], dim=1)
        return ffn_layer(aggr_outputx), aggr_outputx

    @staticmethod
    def _remove_rev_bond_message(original_message, aggr_message, b2revb):
        rev_message = original_message[b2revb]
        return aggr_message - rev_message

    def _atom_bond_transform(
            self,
            to_atom=True,  # False: to bond
            atomwise_input=None,
            bondwise_input=None,
            original_f_atoms=None,
            original_f_bonds=None,
            a2a=None,
            a2b=None,
            b2a=None,
            b2revb=None):
        """Transfer the output of atom/bond multi-head attention to the final atom/bond output.
        """

        if to_atom:
            # atom input to atom output
            atomwise_input, _ = self._pointwise_feed_forward_to_atom_embedding(
                atomwise_input, original_f_atoms, a2a, self.ffn_atom_from_atom)
            atom_in_atom_out = self.atom_from_atom_sublayer(
                None, atomwise_input)
            # bond to atom
            bondwise_input, _ = self._pointwise_feed_forward_to_atom_embedding(
                bondwise_input, original_f_atoms, a2b, self.ffn_atom_from_bond)
            bond_in_atom_out = self.atom_from_bond_sublayer(
                None, bondwise_input)
            return atom_in_atom_out, bond_in_atom_out
        else:  # to bond embeddings
            # atom input to bond output
            atom_list_for_bond = torch.cat([b2a.unsqueeze(dim=1), a2a[b2a]],
                                           dim=1)
            atomwise_input, _ = self._pointwise_feed_forward_to_bond_embedding(
                atomwise_input, original_f_bonds, atom_list_for_bond,
                b2a[b2revb], self.ffn_bond_from_atom)
            atom_in_bond_out = self.bond_from_atom_sublayer(
                None, atomwise_input)
            # bond input to bond output
            bond_list_for_bond = a2b[b2a]
            bondwise_input, _ = self._pointwise_feed_forward_to_bond_embedding(
                bondwise_input, original_f_bonds, bond_list_for_bond, b2revb,
                self.ffn_bond_from_bond)
            bond_in_bond_out = self.bond_from_bond_sublayer(
                None, bondwise_input)
            return atom_in_bond_out, bond_in_bond_out

    def forward(self, batch):
        """Forward layer

        Parameters
        ----------
        batch: Tuple
            A tuple of tensors representing grover attributes

        Returns
        -------
        embeddings: Tuple[Tuple[torch.Tensor, torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]
            Embeddings for atom generated from hidden state of nodes and bonds and embeddings of bond generated from hidden states of nodes and bond.
        """

        f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, a2a = batch

        node_batch = f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, a2a
        edge_batch = f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, a2a

        original_f_atoms, original_f_bonds = f_atoms, f_bonds

        for nb in self.node_blocks:  # atom messages. Multi-headed attention
            node_batch = nb(node_batch)
        for eb in self.edge_blocks:  # bond messages. Multi-headed attention
            edge_batch = eb(edge_batch)

        atom_output, _, _, _, _, _, _, _ = node_batch  # atom hidden states
        _, bond_output, _, _, _, _, _, _ = edge_batch  # bond hidden states

        atom_embeddings = self._atom_bond_transform(
            to_atom=True,  # False: to bond
            atomwise_input=atom_output,
            bondwise_input=bond_output,
            original_f_atoms=original_f_atoms,
            original_f_bonds=original_f_bonds,
            a2a=a2a,
            a2b=a2b,
            b2a=b2a,
            b2revb=b2revb)

        bond_embeddings = self._atom_bond_transform(
            to_atom=False,  # False: to bond
            atomwise_input=atom_output,
            bondwise_input=bond_output,
            original_f_atoms=original_f_atoms,
            original_f_bonds=original_f_bonds,
            a2a=a2a,
            a2b=a2b,
            b2a=b2a,
            b2revb=b2revb)

        return ((atom_embeddings[0], bond_embeddings[0]), (atom_embeddings[1],
                                                           bond_embeddings[1]))



import numpy as np
from deepchem.feat import Featurizer
from deepchem.data import ImageDataset
from typing import List


class PileupFeaturizer(Featurizer):
    """
    Generates Pileup Images from haplotype windows.

    PileupFeaturizer generates pileup images from DNA sequence
    data in a specific genomic window. Each image encodes sequence
    information across multiple channels, representing features
    like base identity, base quality, mapping quality, strand
    orientation, variant support, and differences from the
    reference sequence. This featurizer decodes one-hot encoded
    reference sequences, aligns haplotypes, and calculates intensity
    values for each feature in the pileup image. The generated images
    are used for downstream tasks, such as variant calling.

    Examples
    --------
    >>> from deepchem.feat import RealignerFeaturizer, PileupFeaturizer
    >>> bamfile_path = 'deepchem/data/tests/example.bam'
    >>> reference_path = 'deepchem/data/tests/sample.fa'
    >>> realign = RealignerFeaturizer()
    >>> datapoint = (bamfile_path, reference_path)
    >>> features = realign.featurize([datapoint])
    >>> windows_haplotypes = features[0]
    >>> pileup_feat = PileupFeaturizer()
    >>> datapoint = (windows_haplotypes, reference_path, 299, 299, 6)
    >>> features = pileup_feat.featurize([datapoint])

    Note
    ----
    This class requires pysam to be installed. Pysam can be used with
    Linux or MacOS X. To use Pysam on Windows, use Windows Subsystem for
    Linux(WSL).

    """

    def decode_one_hot(self,
                       one_hot_vector: List[np.ndarray],
                       charset: List[str] = ["A", "C", "T", "G", "N"]) -> str:
        """
        Decode a one-hot encoded sequence into a string of
        nucleotides.

        This function is not used and retained only for backward
        compatibility with previous versions of the code.

        Parameters
        ----------
        one_hot_vector : List[np.ndarray]
            List of one-hot encoded vectors.
        charset : Optional[List[str]]
            List of characters corresponding to the encoding.
            Default is ["A", "C", "T", "G", "N"].

        Returns
        -------
        str
            Decoded sequence as a string.
        """
        decoded_seq = []
        for vector in one_hot_vector:
            idx = np.argmax(vector)
            decoded_seq.append(charset[idx])
        return ''.join(decoded_seq)

    def _featurize(self, datapoint):
        """
        Featurizes a datapoint by generating pileup images.

        Parameters
        ----------

        datapoint : Tuple[List[Any], str]
            A tuple containing haplotypes, reference file path,
            height, width, and num_channels.

        Returns
        -------

        ImageDataset
            An ImageDataset containing the images and labels.

        """
        windows_haplotypes = datapoint[0]
        reference_file_path = datapoint[1]

        from deepchem.data import FASTALoader

        # Load the reference FASTA file with raw fasta sequences
        fasta_loader = FASTALoader(None, False, False, True)
        fasta_dataset = fasta_loader.create_dataset(reference_file_path)

        # The X array contains [name, sequence] pairs
        name_seq_pairs = fasta_dataset.X

        # Create dictionary directly from the pairs
        reference_seq_dict = {pair[0]: pair[1] for pair in name_seq_pairs}

        def base_to_intensity(base):
            if base == "A":
                return 0.25
            elif base == "C":
                return 0.5
            elif base == "G":
                return 0.75
            elif base == "T":
                return 1.0
            else:
                return 0

        def get_base_quality_intensity(quality):
            return quality / 40.0

        def get_mapping_quality_intensity(mapping_quality):
            return mapping_quality / 60.0

        def get_strand_intensity(is_reverse):
            return 1.0 if is_reverse else 0.0

        def get_supports_variant_intensity(read, haplotype):
            read_bases = read[1]
            query_length = read[2]
            ref_bases = haplotype[:query_length]
            mismatch = False
            for rb, hb in zip(read_bases, ref_bases):
                if rb != hb:
                    mismatch = True
                    break
            return 1.0 if mismatch else 0.5

        def get_diff_from_ref_intensity(base, ref_base):
            return 1.0 if base != ref_base else 0.25

        height = datapoint[2]  # 299
        width = datapoint[3]  # 299
        num_channels = datapoint[4]  # 6

        images = []
        labels = []

        for window in windows_haplotypes:
            chrom, start, end = window['span']
            haplotypes = window['haplotypes']
            realigned_reads = window['realigned_reads']

            if not haplotypes:
                continue

            variant_pos = (start + end) // 2  # Variant position in the center
            start = max(0, variant_pos - height // 2)
            end = start + height

            ref_sequence = reference_seq_dict[chrom][start:end + 1]
            image = np.zeros((width, height, num_channels))

            for i, read in enumerate(realigned_reads):
                if i >= width:
                    break
                for j in range(read[2]):
                    ref_pos = read[4] + j - start
                    if ref_pos < 0 or ref_pos >= height or ref_pos >= len(
                            ref_sequence):
                        continue

                    base = read[1][j]
                    base_quality = read[8][j]
                    mapping_quality = read[6]
                    is_reverse = read[7]
                    supports_variant = get_supports_variant_intensity(
                        read, haplotypes[0])
                    diff_from_ref = get_diff_from_ref_intensity(
                        base, ref_sequence[ref_pos])

                    image[i, ref_pos, 0] = base_to_intensity(base)
                    image[i, ref_pos,
                          1] = get_base_quality_intensity(base_quality)
                    image[i, ref_pos,
                          2] = get_mapping_quality_intensity(mapping_quality)
                    image[i, ref_pos, 3] = get_strand_intensity(is_reverse)
                    image[i, ref_pos, 4] = supports_variant
                    image[i, ref_pos, 5] = diff_from_ref

            images.append(image)  # Store the generated image
            labels.append(1)

        # Convert the image list to numpy array
        images_array = np.array(images)
        labels_array = np.array(labels)

        # Create ImageDataset
        image_dataset = ImageDataset(X=images_array, y=labels_array)

        return image_dataset



"""
Gathers all splitters in one place for convenient imports
"""
# flake8: noqa

# basic splitter
from deepchem.splits.splitters import Splitter
from deepchem.splits.splitters import RandomSplitter
from deepchem.splits.splitters import RandomStratifiedSplitter
from deepchem.splits.splitters import RandomGroupSplitter
from deepchem.splits.splitters import SingletaskStratifiedSplitter
from deepchem.splits.splitters import IndexSplitter
from deepchem.splits.splitters import SpecifiedSplitter

# molecule splitter
from deepchem.splits.splitters import ScaffoldSplitter
from deepchem.splits.splitters import MolecularWeightSplitter
from deepchem.splits.splitters import MaxMinSplitter
from deepchem.splits.splitters import FingerprintSplitter
from deepchem.splits.splitters import ButinaSplitter

# other splitter
from deepchem.splits.task_splitter import merge_fold_datasets
from deepchem.splits.task_splitter import TaskSplitter

#################################################################
# Removed API
#################################################################

import logging

logger = logging.getLogger(__name__)


class IndiceSplitter:

    def __init__(self, valid_indices=None, test_indices=None):
        raise ImportError(
            "IndiceSplitter was renamed to SpecifiedSplitter.\n"
            "Please use SpecifiedSplitter instead of IndiceSplitter.")



# Requriments - transformers, tokenizers
# Right now, the Smiles Tokenizer uses an exiesting vocab file from rxnfp that is fairly comprehensive and from the USPTO dataset.
# The vocab may be expanded in the near future

import collections
import os
import re
import importlib.resources
from typing import List, Optional
from transformers import BertTokenizer
from logging import getLogger

logger = getLogger(__name__)
"""
SMI_REGEX_PATTERN: str
SMILES regex pattern for tokenization. Designed by Schwaller et. al.

References
----------
.. [1]  Philippe Schwaller, Teodoro Laino, Théophile Gaudin, Peter Bolgar, Christopher A. Hunter, Costas Bekas, and Alpha A. Lee
ACS Central Science 2019 5 (9): Molecular Transformer: A Model for Uncertainty-Calibrated Chemical Reaction Prediction
1572-1583 DOI: 10.1021/acscentsci.9b00576
"""

SMI_REGEX_PATTERN = r"""(\[[^\]]+]|Br?|Cl?|N|O|S|P|F|I|b|c|n|o|s|p|\(|\)|\.|=|#|-|\+|\\|\/|:|~|@|\?|>>?|\*|\$|\%[0-9]{2}|[0-9])"""

# add vocab_file dict
VOCAB_FILES_NAMES = {"vocab_file": "vocab.txt"}


def get_default_tokenizer():
    default_vocab_path = (importlib.resources.resource_filename(
        "deepchem", "feat/tests/vocab.txt"))
    return SmilesTokenizer(default_vocab_path)


class SmilesTokenizer(BertTokenizer):
    """
    Creates the SmilesTokenizer class. The tokenizer heavily inherits from the BertTokenizer
    implementation found in Huggingface's transformers library. It runs a WordPiece tokenization
    algorithm over SMILES strings using the tokenisation SMILES regex developed by Schwaller et. al.

    Please see https://github.com/huggingface/transformers
    and https://github.com/rxn4chemistry/rxnfp for more details.

    Examples
    --------
    >>> from deepchem.feat.smiles_tokenizer import SmilesTokenizer
    >>> current_dir = os.path.dirname(os.path.realpath(__file__))
    >>> vocab_path = os.path.join(current_dir, 'tests/data', 'vocab.txt')
    >>> tokenizer = SmilesTokenizer(vocab_path)
    >>> print(tokenizer.encode("CC(=O)OC1=CC=CC=C1C(=O)O"))
    [12, 16, 16, 17, 22, 19, 18, 19, 16, 20, 22, 16, 16, 22, 16, 16, 22, 16, 20, 16, 17, 22, 19, 18, 19, 13]


    References
    ----------
    .. [1] Schwaller, Philippe; Probst, Daniel; Vaucher, Alain C.; Nair, Vishnu H; Kreutter, David;
        Laino, Teodoro; et al. (2019): Mapping the Space of Chemical Reactions using Attention-Based Neural
        Networks. ChemRxiv. Preprint. https://doi.org/10.26434/chemrxiv.9897365.v3

    Note
    ----
    This class requires huggingface's transformers and tokenizers libraries to be installed.
    """
    vocab_files_names = VOCAB_FILES_NAMES

    def __init__(
        self,
        vocab_file: str = '',
        # unk_token="[UNK]",
        # sep_token="[SEP]",
        # pad_token="[PAD]",
        # cls_token="[CLS]",
        # mask_token="[MASK]",
        **kwargs):
        """Constructs a SmilesTokenizer.

        Parameters
        ----------
        vocab_file: str
            Path to a SMILES character per line vocabulary file.
            Default vocab file is found in deepchem/feat/tests/data/vocab.txt
        """

        super().__init__(vocab_file, **kwargs)

        if not os.path.isfile(vocab_file):
            raise ValueError(
                "Can't find a vocab file at path '{}'.".format(vocab_file))
        self.vocab = load_vocab(vocab_file)
        self.highest_unused_index = max([
            i for i, v in enumerate(self.vocab.keys())
            if v.startswith("[unused")
        ])
        self.ids_to_tokens = collections.OrderedDict([
            (ids, tok) for tok, ids in self.vocab.items()
        ])
        self.basic_tokenizer = BasicSmilesTokenizer()

    @property
    def vocab_size(self):
        return len(self.vocab)

    @property
    def vocab_list(self):
        return list(self.vocab.keys())

    def _tokenize(  # type: ignore
            self, text: str, max_seq_length: int = 512, **kwargs):
        """Tokenize a string into a list of tokens.

        Parameters
        ----------
        text: str
            Input string sequence to be tokenized.
        """

        max_len_single_sentence = max_seq_length - 2
        split_tokens = [
            token for token in self.basic_tokenizer.tokenize(text)
            [:max_len_single_sentence]
        ]
        return split_tokens

    def _convert_token_to_id(self, token: str):
        """Converts a token (str/unicode) in an id using the vocab.

        Parameters
        ----------
        token: str
            String token from a larger sequence to be converted to a numerical id.
        """

        return self.vocab.get(token, self.vocab.get(self.unk_token))

    def _convert_id_to_token(self, index: int):
        """Converts an index (integer) in a token (string/unicode) using the vocab.

        Parameters
        ----------
        index: int
            Integer index to be converted back to a string-based token as part of a larger sequence.
        """

        return self.ids_to_tokens.get(index, self.unk_token)

    def convert_tokens_to_string(self, tokens: List[str]):
        """Converts a sequence of tokens (string) in a single string.

        Parameters
        ----------
        tokens: List[str]
            List of tokens for a given string sequence.

        Returns
        -------
        out_string: str
            Single string from combined tokens.
        """

        out_string: str = " ".join(tokens).replace(" ##", "").strip()
        return out_string

    def add_special_tokens_ids_single_sequence(self,
                                               token_ids: List[Optional[int]]):
        """Adds special tokens to the a sequence for sequence classification tasks.

        A BERT sequence has the following format: [CLS] X [SEP]

        Parameters
        ----------
        token_ids: list[int]
            list of tokenized input ids. Can be obtained using the encode or encode_plus methods.
        """

        return [self.cls_token_id] + token_ids + [self.sep_token_id]

    def add_special_tokens_single_sequence(self, tokens: List[str]):
        """Adds special tokens to the a sequence for sequence classification tasks.
        A BERT sequence has the following format: [CLS] X [SEP]

        Parameters
        ----------
        tokens: List[str]
            List of tokens for a given string sequence.
        """
        return [self.cls_token] + tokens + [self.sep_token]

    def add_special_tokens_ids_sequence_pair(
            self, token_ids_0: List[Optional[int]],
            token_ids_1: List[Optional[int]]) -> List[Optional[int]]:
        """Adds special tokens to a sequence pair for sequence classification tasks.
        A BERT sequence pair has the following format: [CLS] A [SEP] B [SEP]

        Parameters
        ----------
        token_ids_0: List[int]
            List of ids for the first string sequence in the sequence pair (A).
        token_ids_1: List[int]
            List of tokens for the second string sequence in the sequence pair (B).
        """

        sep = [self.sep_token_id]
        cls = [self.cls_token_id]

        return cls + token_ids_0 + sep + token_ids_1 + sep

    def add_padding_tokens(self,
                           token_ids: List[Optional[int]],
                           length: int,
                           right: bool = True) -> List[Optional[int]]:
        """Adds padding tokens to return a sequence of length max_length.
        By default padding tokens are added to the right of the sequence.

        Parameters
        ----------
        token_ids: list[optional[int]]
            list of tokenized input ids. Can be obtained using the encode or encode_plus methods.
        length: int
            TODO
        right: bool, default True
            TODO

        Returns
        -------
        List[int]
            TODO
        """
        padding = [self.pad_token_id] * (length - len(token_ids))

        if right:
            return token_ids + padding
        else:
            return padding + token_ids

    def save_vocabulary(
        self,
        save_directory: str,
        filename_prefix: Optional[str] = None
    ):  # -> Tuple[str]: doctest issue raised with this return type annotation
        """Save the tokenizer vocabulary to a file.

        Parameters
        ----------
        vocab_path: obj: str
            The directory in which to save the SMILES character per line vocabulary file.
            Default vocab file is found in deepchem/feat/tests/data/vocab.txt

        Returns
        -------
        vocab_file: Tuple
            Paths to the files saved.
            typle with string to a SMILES character per line vocabulary file.
            Default vocab file is found in deepchem/feat/tests/data/vocab.txt
        """
        index = 0
        if os.path.isdir(save_directory):
            vocab_file = os.path.join(save_directory,
                                      VOCAB_FILES_NAMES["vocab_file"])
        else:
            vocab_file = save_directory
        with open(vocab_file, "w", encoding="utf-8") as writer:
            for token, token_index in sorted(self.vocab.items(),
                                             key=lambda kv: kv[1]):
                if index != token_index:
                    logger.warning(
                        "Saving vocabulary to {}: vocabulary indices are not consecutive."
                        " Please check that the vocabulary is not corrupted!".
                        format(vocab_file))
                    index = token_index
                writer.write(token + "\n")
                index += 1
        return (vocab_file,)


class BasicSmilesTokenizer(object):
    """
    Run basic SMILES tokenization using a regex pattern developed by Schwaller et. al.
    This tokenizer is to be used when a tokenizer that does not require the transformers library by HuggingFace is required.

    Examples
    --------
    >>> from deepchem.feat.smiles_tokenizer import BasicSmilesTokenizer
    >>> tokenizer = BasicSmilesTokenizer()
    >>> print(tokenizer.tokenize("CC(=O)OC1=CC=CC=C1C(=O)O"))
    ['C', 'C', '(', '=', 'O', ')', 'O', 'C', '1', '=', 'C', 'C', '=', 'C', 'C', '=', 'C', '1', 'C', '(', '=', 'O', ')', 'O']


    References
    ----------
    .. [1] Philippe Schwaller, Teodoro Laino, Théophile Gaudin, Peter Bolgar, Christopher A. Hunter, Costas Bekas, and Alpha A. Lee
        ACS Central Science 2019 5 (9): Molecular Transformer: A Model for Uncertainty-Calibrated Chemical Reaction Prediction
        1572-1583 DOI: 10.1021/acscentsci.9b00576
    """

    def __init__(self, regex_pattern: str = SMI_REGEX_PATTERN):
        """Constructs a BasicSMILESTokenizer.

        Parameters
        ----------
        regex: string
            SMILES token regex
        """
        self.regex_pattern = regex_pattern
        self.regex = re.compile(self.regex_pattern)

    def tokenize(self, text):
        """Basic Tokenization of a SMILES.
        """
        tokens = [token for token in self.regex.findall(text)]
        return tokens


def load_vocab(vocab_file):
    """Loads a vocabulary file into a dictionary."""
    vocab = collections.OrderedDict()
    with open(vocab_file, "r", encoding="utf-8") as reader:
        tokens = reader.readlines()
    for index, token in enumerate(tokens):
        token = token.rstrip("\n")
        vocab[token] = index
    return vocab



import torch
import numpy as np
from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.models.torch_models.layers import CNNModule
from deepchem.models.losses import L2Loss
from deepchem.metrics import to_one_hot

from typing import List, Union, Callable, Optional
from deepchem.utils.typing import OneOrMany, ActivationFn, LossFn


class CNN(TorchModel):
    """A 1, 2, or 3 dimensional convolutional network for either regression or classification.

    The network consists of the following sequence of layers:

    - A configurable number of convolutional layers
    - A global pooling layer (either max pool or average pool)
    - A final fully connected layer to compute the output

    It optionally can compose the model from pre-activation residual blocks, as
    described in https://arxiv.org/abs/1603.05027, rather than a simple stack of
    convolution layers.  This often leads to easier training, especially when using a
    large number of layers.  Note that residual blocks can only be used when
    successive layers have the same output shape.  Wherever the output shape changes, a
    simple convolution layer will be used even if residual=True.

    Examples
    --------
    >>> import deepchem as dc
    >>> n_samples = 10
    >>> n_features = 3
    >>> n_tasks = 1
    >>> np.random.seed(123)
    >>> X = np.random.rand(n_samples, 10, n_features)
    >>> y = np.random.randint(2, size=(n_samples, n_tasks)).astype(np.float32)
    >>> dataset: dc.data.Dataset = dc.data.NumpyDataset(X, y)
    >>> regression_metric = dc.metrics.Metric(dc.metrics.mean_squared_error)
    >>> model = CNN(n_tasks, n_features, dims=1, kernel_size=3, mode='regression')
    >>> avg_loss = model.fit(dataset, nb_epoch=10)

    """

    def __init__(self,
                 n_tasks: int,
                 n_features: int,
                 dims: int,
                 layer_filters: List[int] = [100],
                 kernel_size: OneOrMany[int] = 5,
                 strides: OneOrMany[int] = 1,
                 weight_init_stddevs: OneOrMany[float] = 0.02,
                 bias_init_consts: OneOrMany[float] = 1.0,
                 weight_decay_penalty: float = 0.0,
                 weight_decay_penalty_type: str = 'l2',
                 dropouts: OneOrMany[float] = 0.5,
                 activation_fns: OneOrMany[ActivationFn] = 'relu',
                 pool_type: str = 'max',
                 mode: str = 'classification',
                 n_classes: int = 2,
                 uncertainty: bool = False,
                 residual: bool = False,
                 padding: Union[int, str] = 'valid',
                 **kwargs) -> None:
        """TorchModel wrapper for CNN

        Parameters
        ----------
        n_tasks: int
            number of tasks
        n_features: int
            number of features
        dims: int
            the number of dimensions to apply convolutions over (1, 2, or 3)
        layer_filters: list
            the number of output filters for each convolutional layer in the network.
            The length of this list determines the number of layers.
        kernel_size: int, tuple, or list
            a list giving the shape of the convolutional kernel for each layer.  Each
            element may be either an int (use the same kernel width for every dimension)
            or a tuple (the kernel width along each dimension).  Alternatively this may
            be a single int or tuple instead of a list, in which case the same kernel
            shape is used for every layer.
        strides: int, tuple, or list
            a list giving the stride between applications of the  kernel for each layer.
            Each element may be either an int (use the same stride for every dimension)
            or a tuple (the stride along each dimension).  Alternatively this may be a
            single int or tuple instead of a list, in which case the same stride is
            used for every layer.
        weight_init_stddevs: list or float
            the standard deviation of the distribution to use for weight initialization
            of each layer.  The length of this list should equal len(layer_filters)+1,
            where the final element corresponds to the dense layer.  Alternatively this
            may be a single value instead of a list, in which case the same value is used
            for every layer.
        bias_init_consts: list or float
            the value to initialize the biases in each layer to.  The length of this
            list should equal len(layer_filters)+1, where the final element corresponds
            to the dense layer.  Alternatively this may be a single value instead of a
            list, in which case the same value is used for every layer.
        weight_decay_penalty: float
            the magnitude of the weight decay penalty to use
        weight_decay_penalty_type: str
            the type of penalty to use for weight decay, either 'l1' or 'l2'
        dropouts: list or float
            the dropout probability to use for each layer.  The length of this list should equal len(layer_filters).
            Alternatively this may be a single value instead of a list, in which case the same value is used for every layer
        activation_fns: str or list
            the torch activation function to apply to each layer. The length of this list should equal
            len(layer_filters).  Alternatively this may be a single value instead of a list, in which case the
            same value is used for every layer, 'relu' by default
        pool_type: str
            the type of pooling layer to use, either 'max' or 'average'
        mode: str
            Either 'classification' or 'regression'
        n_classes: int
            the number of classes to predict (only used in classification mode)
        uncertainty: bool
            if True, include extra outputs and loss terms to enable the uncertainty
            in outputs to be predicted
        residual: bool
            if True, the model will be composed of pre-activation residual blocks instead
            of a simple stack of convolutional layers.
        padding: str, int or tuple
            the padding to use for convolutional layers, either 'valid' or 'same'
        """
        self.mode = mode
        self.n_classes = n_classes
        self.n_tasks = n_tasks

        self.model = CNNModule(n_tasks=n_tasks,
                               n_features=n_features,
                               dims=dims,
                               layer_filters=layer_filters,
                               kernel_size=kernel_size,
                               strides=strides,
                               weight_init_stddevs=weight_init_stddevs,
                               bias_init_consts=bias_init_consts,
                               dropouts=dropouts,
                               activation_fns=activation_fns,
                               pool_type=pool_type,
                               mode=mode,
                               n_classes=n_classes,
                               uncertainty=uncertainty,
                               residual=residual,
                               padding=padding)

        regularization_loss: Optional[Callable]

        if weight_decay_penalty != 0:
            weights = [layer.weight for layer in self.model.layers]
            if weight_decay_penalty_type == 'l1':
                regularization_loss = lambda: weight_decay_penalty * torch.sum(  # noqa: E731
                    torch.stack([torch.abs(w).sum() for w in weights]))
            else:
                regularization_loss = lambda: weight_decay_penalty * torch.sum(  # noqa: E731
                    torch.stack([torch.square(w).sum() for w in weights]))
        else:
            regularization_loss = None

        loss: Union[L2Loss, LossFn]

        if uncertainty:

            def loss(outputs, labels, weights):
                diff = labels[0] - outputs[0]
                return torch.mean(diff**2 / torch.exp(outputs[1]) + outputs[1])

        else:
            loss = L2Loss()

        if self.mode == 'classification':
            output_types = ['prediction', 'loss']
        else:
            if uncertainty:
                output_types = ['prediction', 'variance', 'loss', 'loss']
            else:
                output_types = ["prediction"]

        super(CNN, self).__init__(self.model,
                                  loss=loss,
                                  output_types=output_types,
                                  regularization_loss=regularization_loss,
                                  **kwargs)

    def default_generator(self,
                          dataset,
                          epochs=1,
                          mode='fit',
                          deterministic=True,
                          pad_batches=True):

        for epoch in range(epochs):

            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):

                if self.mode == 'classification':
                    if y_b is not None:
                        y_b = to_one_hot(y_b.flatten(), self.n_classes)\
                            .reshape(-1, self.n_tasks, self.n_classes)

                dropout = np.array(0.) if mode == 'predict' else np.array(1.)

                yield ([X_b, dropout], [y_b], [w_b])



from rdkit import Chem
import numpy as np
import logging
from typing import List, Tuple, Union, Dict, Set, Sequence, Optional
from deepchem.utils.typing import RDKitAtom, RDKitMol, RDKitBond

from deepchem.feat.base_classes import MolecularFeaturizer
from deepchem.feat.graph_data import GraphData
from deepchem.feat.molecule_featurizers.circular_fingerprint import CircularFingerprint
from deepchem.feat.molecule_featurizers.rdkit_descriptors import RDKitDescriptors

from deepchem.utils.molecule_feature_utils import one_hot_encode
from deepchem.utils.molecule_feature_utils import get_atom_total_degree_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_formal_charge_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_total_num_Hs_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_hybridization_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_is_in_aromatic_one_hot

from deepchem.feat.graph_features import bond_features as b_Feats

logger = logging.getLogger(__name__)


class GraphConvConstants(object):
    """
    A class for holding featurization parameters.
    """

    MAX_ATOMIC_NUM = 100
    ATOM_FEATURES: Dict[str, List[int]] = {
        'atomic_num': list(range(MAX_ATOMIC_NUM)),
        'degree': [0, 1, 2, 3, 4, 5],
        'formal_charge': [-1, -2, 1, 2, 0],
        'chiral_tag': [0, 1, 2, 3],
        'num_Hs': [0, 1, 2, 3, 4]
    }
    ATOM_FEATURES_HYBRIDIZATION: List[str] = [
        "SP", "SP2", "SP3", "SP3D", "SP3D2"
    ]
    # Dimension of atom feature vector
    ATOM_FDIM = sum(len(choices) + 1 for choices in ATOM_FEATURES.values()
                   ) + len(ATOM_FEATURES_HYBRIDIZATION) + 1 + 2
    # len(choices) +1 and len(ATOM_FEATURES_HYBRIDIZATION) +1 to include room for unknown set
    # + 2 at end for is_in_aromatic and mass
    BOND_FDIM = 14

    # dictionary of available feature generators
    FEATURE_GENERATORS: Dict[str, MolecularFeaturizer] = {
        "morgan":
            CircularFingerprint(radius=2, size=2048, sparse=False),
        "morgan_count":
            CircularFingerprint(radius=2,
                                size=2048,
                                sparse=False,
                                is_counts_based=True),
        "rdkit_desc":
            RDKitDescriptors(use_bcut2d=False),
        "rdkit_desc_normalized":
            RDKitDescriptors(use_bcut2d=False, is_normalized=True)
    }


def get_atomic_num_one_hot(atom: RDKitAtom,
                           allowable_set: List[int],
                           include_unknown_set: bool = True) -> List[float]:
    """Get a one-hot feature about atomic number of the given atom.

    Parameters
    ---------
    atom: RDKitAtom
        RDKit atom object
    allowable_set: List[int]
        The range of atomic numbers to consider.
    include_unknown_set: bool, default False
        If true, the index of all types not in `allowable_set` is `len(allowable_set)`.

    Returns
    -------
    List[float]
        A one-hot vector of atomic number of the given atom.
        If `include_unknown_set` is False, the length is `len(allowable_set)`.
        If `include_unknown_set` is True, the length is `len(allowable_set) + 1`.

    """
    return one_hot_encode(atom.GetAtomicNum() - 1, allowable_set,
                          include_unknown_set)


def get_atom_chiral_tag_one_hot(
        atom: RDKitAtom,
        allowable_set: List[int],
        include_unknown_set: bool = True) -> List[float]:
    """Get a one-hot feature about chirality of the given atom.

    Parameters
    ---------
    atom: RDKitAtom
        RDKit atom object
    allowable_set: List[int]
        The list of chirality tags to consider.
    include_unknown_set: bool, default False
        If true, the index of all types not in `allowable_set` is `len(allowable_set)`.

    Returns
    -------
    List[float]
        A one-hot vector of chirality of the given atom.
        If `include_unknown_set` is False, the length is `len(allowable_set)`.
        If `include_unknown_set` is True, the length is `len(allowable_set) + 1`.

    """
    return one_hot_encode(atom.GetChiralTag(), allowable_set,
                          include_unknown_set)


def get_atom_mass(atom: RDKitAtom) -> List[float]:
    """Get vector feature containing downscaled mass of the given atom.

    Parameters
    ---------
    atom: RDKitAtom
        RDKit atom object

    Returns
    -------
    List[float]
        A vector of downscaled mass of the given atom.

    """
    return [atom.GetMass() * 0.01]


def atom_features(
        atom: RDKitAtom,
        functional_groups: Optional[List[int]] = None,
        only_atom_num: bool = False) -> Sequence[Union[bool, int, float]]:
    """Helper method used to compute atom feature vector.

    Deepchem already contains an atom_features function, however we are defining a new one here due to the need to handle features specific to DMPNN.

    Parameters
    ----------
    atom: RDKitAtom
        Atom to compute features on.
    functional_groups: List[int]
        A k-hot vector indicating the functional groups the atom belongs to.
        Default value is None
    only_atom_num: bool
        Toggle to build a feature vector for an atom containing only the atom number information.

    Returns
    -------
    features: Sequence[Union[bool, int, float]]
        A list of atom features.

    Examples
    --------
    >>> import deepchem as dc
    >>> from rdkit import Chem
    >>> mol = Chem.MolFromSmiles('C')
    >>> atom = mol.GetAtoms()[0]
    >>> features = dc.feat.molecule_featurizers.dmpnn_featurizer.atom_features(atom)
    >>> type(features)
    <class 'list'>
    >>> len(features)
    133

    """

    if atom is None:
        features: Sequence[Union[bool, int,
                                 float]] = [0] * GraphConvConstants.ATOM_FDIM

    elif only_atom_num:
        features = []
        features += get_atomic_num_one_hot(
            atom, GraphConvConstants.ATOM_FEATURES['atomic_num'])
        features += [0] * (
            GraphConvConstants.ATOM_FDIM - GraphConvConstants.MAX_ATOMIC_NUM - 1
        )  # set other features to zero

    else:
        features = []
        features += get_atomic_num_one_hot(
            atom, GraphConvConstants.ATOM_FEATURES['atomic_num'])
        features += get_atom_total_degree_one_hot(
            atom, GraphConvConstants.ATOM_FEATURES['degree'])
        features += get_atom_formal_charge_one_hot(
            atom, GraphConvConstants.ATOM_FEATURES['formal_charge'])
        features += get_atom_chiral_tag_one_hot(
            atom, GraphConvConstants.ATOM_FEATURES['chiral_tag'])
        features += get_atom_total_num_Hs_one_hot(
            atom, GraphConvConstants.ATOM_FEATURES['num_Hs'])
        features += get_atom_hybridization_one_hot(
            atom, GraphConvConstants.ATOM_FEATURES_HYBRIDIZATION, True)
        features += get_atom_is_in_aromatic_one_hot(atom)
        features = [int(feature) for feature in features]
        features += get_atom_mass(atom)

        if functional_groups is not None:
            features += functional_groups
    return features


def bond_features(bond: RDKitBond) -> Sequence[Union[bool, int, float]]:
    """wrapper function for bond_features() already available in deepchem, used to compute bond feature vector.

    Parameters
    ----------
    bond: RDKitBond
        Bond to compute features on.

    Returns
    -------
    features: Sequence[Union[bool, int, float]]
        A list of bond features.

    Examples
    --------
    >>> import deepchem as dc
    >>> from rdkit import Chem
    >>> mol = Chem.MolFromSmiles('CC')
    >>> bond = mol.GetBondWithIdx(0)
    >>> b_features = dc.feat.molecule_featurizers.dmpnn_featurizer.bond_features(bond)
    >>> type(b_features)
    <class 'list'>
    >>> len(b_features)
    14

    """
    if bond is None:
        b_features: Sequence[Union[
            bool, int, float]] = [1] + [0] * (GraphConvConstants.BOND_FDIM - 1)

    else:
        b_features = [0] + b_Feats(bond, use_extended_chirality=True)
    return b_features


def map_reac_to_prod(
        mol_reac: RDKitMol,
        mol_prod: RDKitMol) -> Tuple[Dict[int, int], List[int], List[int]]:
    """Function to build a dictionary of mapping atom indices in the reactants to the products.

    Parameters
    ----------
    mol_reac: RDKitMol
    An RDKit molecule of the reactants.

    mol_prod: RDKitMol
    An RDKit molecule of the products.

    Returns
    -------
    mappings: Tuple[Dict[int,int],List[int],List[int]]
    A tuple containing a dictionary of corresponding reactant and product atom indices,
    list of atom ids of product not part of the mapping and
    list of atom ids of reactant not part of the mapping

    """
    only_prod_ids: List[int] = []
    prod_map_to_id: Dict[int, int] = {}
    mapnos_reac: Set[int] = set(
        [atom.GetAtomMapNum() for atom in mol_reac.GetAtoms()])
    for atom in mol_prod.GetAtoms():
        mapno = atom.GetAtomMapNum()
        if (mapno > 0):
            prod_map_to_id[mapno] = atom.GetIdx()
            if (mapno not in mapnos_reac):
                only_prod_ids.append(atom.GetIdx())
        else:
            only_prod_ids.append(atom.GetIdx())
    only_reac_ids: List[int] = []
    reac_id_to_prod_id: Dict[int, int] = {}
    for atom in mol_reac.GetAtoms():
        mapno = atom.GetAtomMapNum()
        if (mapno > 0):
            try:
                reac_id_to_prod_id[atom.GetIdx()] = prod_map_to_id[mapno]
            except KeyError:
                only_reac_ids.append(atom.GetIdx())
        else:
            only_reac_ids.append(atom.GetIdx())
    mappings: Tuple[Dict[int, int], List[int],
                    List[int]] = (reac_id_to_prod_id, only_prod_ids,
                                  only_reac_ids)
    return mappings


def generate_global_features(mol: RDKitMol,
                             features_generators: List[str]) -> np.ndarray:
    """Helper function for generating global features for a RDKit mol based on the given list of feature generators to be used.

    Parameters
    ----------
    mol: RDKitMol
        RDKit molecule to be featurized
    features_generators: List[str]
        List of names of the feature generators to be used featurization

    Returns
    -------
    global_features_array: np.ndarray
        Array of global features

    Examples
    --------
    >>> from rdkit import Chem
    >>> import deepchem as dc
    >>> mol = Chem.MolFromSmiles('C')
    >>> features_generators = ['morgan']
    >>> global_features = dc.feat.molecule_featurizers.dmpnn_featurizer.generate_global_features(mol, features_generators)
    >>> type(global_features)
    <class 'numpy.ndarray'>
    >>> len(global_features)
    2048
    >>> nonzero_features_indices = global_features.nonzero()[0]
    >>> nonzero_features_indices
    array([1264])
    >>> global_features[nonzero_features_indices[0]]
    1.0

    """
    global_features: List[np.ndarray] = []
    available_generators = GraphConvConstants.FEATURE_GENERATORS

    for generator in features_generators:
        if generator in available_generators:
            global_featurizer = available_generators[generator]
            if mol.GetNumHeavyAtoms() > 0:
                global_features.extend(global_featurizer.featurize(mol)[0])
            # for H2
            elif mol.GetNumHeavyAtoms() == 0:
                # not all features are equally long, so used methane as dummy molecule to determine length
                global_features.extend(
                    np.zeros(
                        len(
                            global_featurizer.featurize(
                                Chem.MolFromSmiles('C'))[0])))
        else:
            logger.warning(f"{generator} generator is not available in DMPNN")

    global_features_array: np.ndarray = np.asarray(global_features)

    # Fix nans in features
    replace_token = 0
    global_features_array = np.where(np.isnan(global_features_array),
                                     replace_token, global_features_array)

    return global_features_array


class DMPNNFeaturizer(MolecularFeaturizer):
    """This class is a featurizer for Directed Message Passing Neural Network (D-MPNN) implementation

    The default node(atom) and edge(bond) representations are based on
    `Analyzing Learned Molecular Representations for Property Prediction paper <https://arxiv.org/pdf/1904.01561.pdf>`_.

    The default node representation are constructed by concatenating the following values,
    and the feature length is 133.

    - Atomic num: A one-hot vector of this atom, in a range of first 100 atoms.
    - Degree: A one-hot vector of the degree (0-5) of this atom.
    - Formal charge: Integer electronic charge, -1, -2, 1, 2, 0.
    - Chirality: A one-hot vector of the chirality tag (0-3) of this atom.
    - Number of Hydrogens: A one-hot vector of the number of hydrogens (0-4) that this atom connected.
    - Hybridization: A one-hot vector of "SP", "SP2", "SP3", "SP3D", "SP3D2".
    - Aromatic: A one-hot vector of whether the atom belongs to an aromatic ring.
    - Mass: Atomic mass * 0.01

    The default edge representation are constructed by concatenating the following values,
    and the feature length is 14.

    - Bond type: A one-hot vector of the bond type, "single", "double", "triple", or "aromatic".
    - Same ring: A one-hot vector of whether the atoms in the pair are in the same ring.
    - Conjugated: A one-hot vector of whether this bond is conjugated or not.
    - Stereo: A one-hot vector of the stereo configuration (0-5) of a bond.

    If you want to know more details about features, please check the paper [1]_ and
    utilities in deepchem.utils.molecule_feature_utils.py.

    Examples
    --------
    >>> smiles = ["C1=CC=CN=C1", "C1CCC1"]
    >>> featurizer = DMPNNFeaturizer()
    >>> out = featurizer.featurize(smiles)
    >>> type(out[0])
    <class 'deepchem.feat.graph_data.GraphData'>
    >>> out[0].num_nodes
    6
    >>> out[0].num_node_features
    133
    >>> out[0].node_features.shape
    (6, 133)
    >>> out[0].num_edge_features
    14
    >>> out[0].num_edges
    12
    >>> out[0].edge_features.shape
    (12, 14)

    References
    ----------
    .. [1] Kearnes, Steven, et al. "Molecular graph convolutions: moving beyond fingerprints."
        Journal of computer-aided molecular design 30.8 (2016):595-608.

    Note
    ----
    This class requires RDKit to be installed.

    """

    def __init__(self,
                 features_generators: Optional[List[str]] = None,
                 is_adding_hs: bool = False,
                 use_original_atom_ranks: bool = False):
        """
        Parameters
        ----------
        features_generator: List[str], default None
            List of global feature generators to be used.
        is_adding_hs: bool, default False
            Whether to add Hs or not.
        use_original_atom_ranks: bool, default False
            Whether to use original atom mapping or canonical atom mapping

        """
        self.features_generators = features_generators
        self.is_adding_hs = is_adding_hs
        super().__init__(use_original_atom_ranks)

    def _construct_bond_index(self, datapoint: RDKitMol) -> np.ndarray:
        """
        Construct edge (bond) index

        Parameters
        ----------
        datapoint: RDKitMol
            RDKit mol object.

        Returns
        -------
        edge_index: np.ndarray
            Edge (Bond) index

        """
        src: List[int] = []
        dest: List[int] = []
        for bond in datapoint.GetBonds():
            # add edge list considering a directed graph
            start, end = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
            src += [start, end]
            dest += [end, start]
        return np.asarray([src, dest], dtype=int)

    def _get_bond_features(self, datapoint: RDKitMol) -> np.ndarray:
        """
        Construct bond(edge) features for the given datapoint

        For each bond index, 2 bond feature arrays are added to the main features array,
        for the current bond and its reverse bond respectively.

        Note: This method of generating bond features ensures that the shape of the bond features array
            is always equal to (number of bonds, number of bond features), even if the number of bonds
            is equal to 0.

        Parameters
        ----------
        datapoint: RDKitMol
            RDKit mol object.

        Returns
        -------
        f_bonds: np.ndarray
            Bond features array

        """
        bonds: Chem.rdchem._ROBondSeq = datapoint.GetBonds()

        bond_fdim: int = GraphConvConstants.BOND_FDIM
        number_of_bonds: int = len(
            bonds) * 2  # Note the value is doubled to account for reverse bonds
        f_bonds: np.ndarray = np.empty((number_of_bonds, bond_fdim))

        for index in range(0, number_of_bonds, 2):
            bond_id: int = index // 2
            bond_feature: np.ndarray = np.asarray(bond_features(bonds[bond_id]),
                                                  dtype=float)
            f_bonds[index] = bond_feature  # bond
            f_bonds[index + 1] = bond_feature  # reverse bond
        return f_bonds

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> GraphData:
        """Calculate molecule graph features from RDKit mol object.

        Parameters
        ----------
        datapoint: RDKitMol
            RDKit mol object.

        Returns
        -------
        graph: GraphData
            A molecule graph object with features:
            - node_features: Node feature matrix with shape [num_nodes, num_node_features]
            - edge_index: Graph connectivity in COO format with shape [2, num_edges]
            - edge_features: Edge feature matrix with shape [num_edges, num_edge_features]
            - global_features: Array of global molecular features

        """
        if isinstance(datapoint, Chem.rdchem.Mol):
            if self.is_adding_hs:
                datapoint = Chem.AddHs(datapoint)
        else:
            raise ValueError(
                "Feature field should contain smiles for DMPNN featurizer!")

        # get atom features
        f_atoms: np.ndarray = np.asarray(
            [atom_features(atom) for atom in datapoint.GetAtoms()], dtype=float)

        # get edge(bond) features
        f_bonds: np.ndarray = self._get_bond_features(datapoint)

        # get edge index
        edge_index: np.ndarray = self._construct_bond_index(datapoint)

        # get global features
        global_features: np.ndarray = np.empty(0)
        if self.features_generators is not None:
            global_features = generate_global_features(datapoint,
                                                       self.features_generators)

        return GraphData(node_features=f_atoms,
                         edge_index=edge_index,
                         edge_features=f_bonds,
                         global_features=global_features)



import torch
import torch.nn as nn
import torch.nn.functional as F
from deepchem.models.losses import CategoricalCrossEntropy
from deepchem.models.torch_models import TorchModel
from deepchem.data import Dataset
from deepchem.models.optimizers import RMSProp
from typing import Optional, List, Callable, Any, Union
from deepchem.utils.data_utils import load_from_disk, save_to_disk
from deepchem.utils.typing import LossFn


class InceptionV3(nn.Module):
    """
    InceptionV3 model architecture for image classification.
    """

    def __init__(self,
                 num_classes: int = 1000,
                 aux_logits: bool = True,
                 transform_input: bool = False,
                 in_channels: int = 6,
                 dropout_rate: float = 0.5) -> None:
        super(InceptionV3, self).__init__()
        self.aux_logits = aux_logits

        self.Conv2d_1a_3x3 = BasicConv2d(in_channels,
                                         32,
                                         kernel_size=(3, 3),
                                         stride=(2, 2),
                                         padding=(0, 0))
        self.Conv2d_2a_3x3 = BasicConv2d(32,
                                         32,
                                         kernel_size=(3, 3),
                                         stride=(1, 1),
                                         padding=(0, 0))
        self.Conv2d_2b_3x3 = BasicConv2d(32,
                                         64,
                                         kernel_size=(3, 3),
                                         stride=(1, 1),
                                         padding=(1, 1))
        self.maxpool1 = nn.MaxPool2d((3, 3), (2, 2))
        self.Conv2d_3b_1x1 = BasicConv2d(64,
                                         80,
                                         kernel_size=(1, 1),
                                         stride=(1, 1),
                                         padding=(0, 0))
        self.Conv2d_4a_3x3 = BasicConv2d(80,
                                         192,
                                         kernel_size=(3, 3),
                                         stride=(1, 1),
                                         padding=(0, 0))
        self.maxpool2 = nn.MaxPool2d((3, 3), (2, 2))

        self.Mixed_5b = InceptionA(192, 32)
        self.Mixed_5c = InceptionA(256, 64)
        self.Mixed_5d = InceptionA(288, 64)

        self.Mixed_6a = InceptionB(288)

        self.Mixed_6b = InceptionC(768, 128)
        self.Mixed_6c = InceptionC(768, 160)
        self.Mixed_6d = InceptionC(768, 160)
        self.Mixed_6e = InceptionC(768, 192)

        # Auxiliary classifier (only used during training)
        if aux_logits:
            self.AuxLogits = InceptionAux(768, num_classes)

        self.Mixed_7a = InceptionD(768)
        self.Mixed_7b = InceptionE(1280)
        self.Mixed_7c = InceptionE(2048)

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.dropout = nn.Dropout(dropout_rate, True)
        self.fc = nn.Linear(2048, num_classes)

        for module in self.modules():
            if isinstance(module, nn.Conv2d) or isinstance(module, nn.Linear):
                stddev = float(module.stddev) if hasattr(module,
                                                         "stddev") else 0.1
                torch.nn.init.trunc_normal_(module.weight,
                                            mean=0.0,
                                            std=stddev,
                                            a=-2,
                                            b=2)
            elif isinstance(module, nn.BatchNorm2d):
                nn.init.constant_(module.weight, 1)
                nn.init.constant_(module.bias, 0)

    def forward(self, x: torch.Tensor):
        # N x 6 x 299 x 299
        x = self.Conv2d_1a_3x3(x)
        # N x 32 x 149 x 149
        x = self.Conv2d_2a_3x3(x)
        # N x 32 x 147 x 147
        x = self.Conv2d_2b_3x3(x)
        # N x 64 x 147 x 147
        x = self.maxpool1(x)
        # N x 64 x 73 x 73
        x = self.Conv2d_3b_1x1(x)
        # N x 80 x 73 x 73
        x = self.Conv2d_4a_3x3(x)
        # N x 192 x 71 x 71
        x = self.maxpool2(x)
        # N x 192 x 35 x 35
        x = self.Mixed_5b(x)
        # N x 256 x 35 x 35
        x = self.Mixed_5c(x)
        # N x 288 x 35 x 35
        x = self.Mixed_5d(x)
        # N x 288 x 35 x 35
        x = self.Mixed_6a(x)
        # N x 768 x 17 x 17
        x = self.Mixed_6b(x)
        # N x 768 x 17 x 17
        x = self.Mixed_6c(x)
        # N x 768 x 17 x 17
        x = self.Mixed_6d(x)
        # N x 768 x 17 x 17
        x = self.Mixed_6e(x)

        # Auxiliary output if training
        if self.aux_logits and self.training:
            aux = self.AuxLogits(x)

        # N x 768 x 17 x 17
        x = self.Mixed_7a(x)
        # N x 1280 x 8 x 8
        x = self.Mixed_7b(x)
        # N x 2048 x 8 x 8
        x = self.Mixed_7c(x)
        # N x 2048 x 8 x 8
        # Adaptive average pooling
        x = self.avgpool(x)
        # N x 2048 x 1 x 1
        x = torch.flatten(x, 1)
        # N x 2048 x 1 x 1
        x = self.dropout(x)
        # N x 2048
        x = self.fc(x)
        # N x 3 (num_classes)

        if self.aux_logits and self.training:
            return x, aux
        else:
            return x


# Helper layers
class BasicConv2d(nn.Module):
    """
    A basic convolutional layer with Conv2d, BatchNorm2d, and ReLU activation.

    Examples
    --------
    >>> import torch
    >>> from deepchem.models.torch_models.inceptionv3 import BasicConv2d
    >>> layer = BasicConv2d(6, 32, kernel_size=3, stride=2)
    >>> x = torch.randn(5, 6, 299, 299)
    >>> output = layer(x)
    >>> output.shape
    torch.Size([5, 32, 149, 149])
    """

    def __init__(self, in_channels: int, out_channels: int,
                 **kwargs: Any) -> None:
        super(BasicConv2d, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, bias=False, **kwargs)
        self.bn = nn.BatchNorm2d(out_channels, eps=0.001)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass for the BasicConv2d layer.
        """
        x = self.conv(x)
        x = self.bn(x)
        return F.relu(x, inplace=True)


class InceptionA(nn.Module):
    """
    An InceptionA module as part of the InceptionV3 network architecture.
    This module performs parallel convolutions on the input using 1x1,
    5x5, and double 3x3 kernel sizes, as well as an average pooling layer,
    before concatenating the results along the channel dimension.

    Examples
    --------
    >>> import torch
    >>> from deepchem.models.torch_models.inceptionv3 import InceptionA
    >>> layer = InceptionA(192, pool_features=32)
    >>> x = torch.randn(5, 192, 35, 35)
    >>> output = layer(x)
    >>> output.shape
    torch.Size([5, 256, 35, 35])
    """

    def __init__(
        self,
        in_channels: int,
        pool_features: int,
    ) -> None:
        super(InceptionA, self).__init__()
        self.branch1x1 = BasicConv2d(in_channels,
                                     64,
                                     kernel_size=(1, 1),
                                     stride=(1, 1),
                                     padding=(0, 0))

        self.branch5x5_1 = BasicConv2d(in_channels,
                                       48,
                                       kernel_size=(1, 1),
                                       stride=(1, 1),
                                       padding=(0, 0))
        self.branch5x5_2 = BasicConv2d(48,
                                       64,
                                       kernel_size=(5, 5),
                                       stride=(1, 1),
                                       padding=(2, 2))

        self.branch3x3dbl_1 = BasicConv2d(in_channels,
                                          64,
                                          kernel_size=(1, 1),
                                          stride=(1, 1),
                                          padding=(0, 0))
        self.branch3x3dbl_2 = BasicConv2d(64,
                                          96,
                                          kernel_size=(3, 3),
                                          stride=(1, 1),
                                          padding=(1, 1))
        self.branch3x3dbl_3 = BasicConv2d(96,
                                          96,
                                          kernel_size=(3, 3),
                                          stride=(1, 1),
                                          padding=(1, 1))

        self.branch_pool = BasicConv2d(in_channels,
                                       pool_features,
                                       kernel_size=(1, 1),
                                       stride=(1, 1),
                                       padding=(0, 0))
        # self.avgpool = nn.AvgPool2d((3, 3), (1, 1), (1, 1))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass for the InceptionA module.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape (N, in_channels, H, W).

        Returns
        -------
        torch.Tensor
            Output tensor of shape (N, out_channels, H, W),
            where out_channels is the total number of output channels
            from all branches combined.
        """
        branch1x1 = self.branch1x1(x)

        branch5x5 = self.branch5x5_1(x)
        branch5x5 = self.branch5x5_2(branch5x5)

        branch3x3dbl = self.branch3x3dbl_1(x)
        branch3x3dbl = self.branch3x3dbl_2(branch3x3dbl)
        branch3x3dbl = self.branch3x3dbl_3(branch3x3dbl)

        branch_pool = F.avg_pool2d(x, kernel_size=3, stride=1, padding=1)
        branch_pool = self.branch_pool(branch_pool)

        out = [branch1x1, branch5x5, branch3x3dbl, branch_pool]
        output = torch.cat(out, 1)

        return output


class InceptionB(nn.Module):
    """
    An InceptionB module as part of the InceptionV3 network architecture.
    This module performs parallel operations on the input, including a
    3x3 convolution, a double 3x3 convolution, and a max-pooling layer,
    before concatenating the results along the channel dimension.

    Examples
    --------
    >>> import torch
    >>> from deepchem.models.torch_models.inceptionv3 import InceptionB
    >>> layer = InceptionB(288)
    >>> x = torch.randn(5, 288, 35, 35)
    >>> output = layer(x)
    >>> output.shape
    torch.Size([5, 768, 17, 17])
    """

    def __init__(
        self,
        in_channels: int,
    ) -> None:
        super(InceptionB, self).__init__()

        self.branch3x3 = BasicConv2d(in_channels,
                                     384,
                                     kernel_size=(3, 3),
                                     stride=(2, 2),
                                     padding=(0, 0))

        self.branch3x3dbl_1 = BasicConv2d(in_channels,
                                          64,
                                          kernel_size=(1, 1),
                                          stride=(1, 1),
                                          padding=(0, 0))
        self.branch3x3dbl_2 = BasicConv2d(64,
                                          96,
                                          kernel_size=(3, 3),
                                          stride=(1, 1),
                                          padding=(1, 1))
        self.branch3x3dbl_3 = BasicConv2d(96,
                                          96,
                                          kernel_size=(3, 3),
                                          stride=(2, 2),
                                          padding=(0, 0))

        # self.maxpool = nn.MaxPool2d((3, 3), (2, 2))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass for the InceptionB module.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape (N, in_channels, H, W).

        Returns
        -------
        torch.Tensor
            Output tensor of shape (N, out_channels, H, W),
            where out_channels is the total number of output channels
            from all branches combined.
        """
        branch3x3 = self.branch3x3(x)

        branch3x3dbl = self.branch3x3dbl_1(x)
        branch3x3dbl = self.branch3x3dbl_2(branch3x3dbl)
        branch3x3dbl = self.branch3x3dbl_3(branch3x3dbl)

        branch_pool = F.max_pool2d(x, kernel_size=3, stride=2)

        out = [branch3x3, branch3x3dbl, branch_pool]
        output = torch.cat(out, 1)

        return output


class InceptionC(nn.Module):
    """
    An InceptionC module as part of the InceptionV3 network
    architecture. This module performs parallel operations on the
    input, including 1x1, 7x7, double 7x7 convolutions, and an
    average pooling layer, before concatenating the results along
    the channel dimension.

    Examples
    --------
    >>> import torch
    >>> from deepchem.models.torch_models.inceptionv3 import InceptionC
    >>> layer = InceptionC(768, channels_7x7=128)
    >>> x = torch.randn(5, 768, 17, 17)
    >>> output = layer(x)
    >>> output.shape
    torch.Size([5, 768, 17, 17])
    """

    def __init__(
        self,
        in_channels: int,
        channels_7x7: int,
    ) -> None:
        super(InceptionC, self).__init__()

        self.branch1x1 = BasicConv2d(in_channels, 192, kernel_size=1)

        self.branch7x7_1 = BasicConv2d(in_channels,
                                       channels_7x7,
                                       kernel_size=(1, 1),
                                       stride=(1, 1),
                                       padding=(0, 0))
        self.branch7x7_2 = BasicConv2d(channels_7x7,
                                       channels_7x7,
                                       kernel_size=(1, 7),
                                       stride=(1, 1),
                                       padding=(0, 3))
        self.branch7x7_3 = BasicConv2d(channels_7x7,
                                       192,
                                       kernel_size=(7, 1),
                                       stride=(1, 1),
                                       padding=(3, 0))

        self.branch7x7dbl_1 = BasicConv2d(in_channels,
                                          channels_7x7,
                                          kernel_size=1)
        self.branch7x7dbl_2 = BasicConv2d(channels_7x7,
                                          channels_7x7,
                                          kernel_size=(7, 1),
                                          stride=(1, 1),
                                          padding=(3, 0))
        self.branch7x7dbl_3 = BasicConv2d(channels_7x7,
                                          channels_7x7,
                                          kernel_size=(1, 7),
                                          stride=(1, 1),
                                          padding=(0, 3))
        self.branch7x7dbl_4 = BasicConv2d(channels_7x7,
                                          channels_7x7,
                                          kernel_size=(7, 1),
                                          stride=(1, 1),
                                          padding=(3, 0))
        self.branch7x7dbl_5 = BasicConv2d(channels_7x7,
                                          192,
                                          kernel_size=(1, 7),
                                          stride=(1, 1),
                                          padding=(0, 3))

        # self.avgpool = nn.AvgPool2d((3, 3), (1, 1), (1, 1))
        self.branch_pool = BasicConv2d(in_channels,
                                       192,
                                       kernel_size=(1, 1),
                                       stride=(1, 1),
                                       padding=(0, 0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass for the InceptionC module.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape (N, in_channels, H, W).

        Returns
        -------
        torch.Tensor
            Output tensor of shape (N, out_channels, H, W),
            where out_channels is the total number of output channels
            from all branches combined.
        """
        branch1x1 = self.branch1x1(x)

        branch7x7 = self.branch7x7_1(x)
        branch7x7 = self.branch7x7_2(branch7x7)
        branch7x7 = self.branch7x7_3(branch7x7)

        branch7x7dbl = self.branch7x7dbl_1(x)
        branch7x7dbl = self.branch7x7dbl_2(branch7x7dbl)
        branch7x7dbl = self.branch7x7dbl_3(branch7x7dbl)
        branch7x7dbl = self.branch7x7dbl_4(branch7x7dbl)
        branch7x7dbl = self.branch7x7dbl_5(branch7x7dbl)

        branch_pool = F.avg_pool2d(x, kernel_size=3, stride=1, padding=1)
        branch_pool = self.branch_pool(branch_pool)

        out = [branch1x1, branch7x7, branch7x7dbl, branch_pool]
        output = torch.cat(out, 1)

        return output


class InceptionD(nn.Module):
    """
    An InceptionD module as part of the InceptionV3 network
    architecture. This module performs parallel operations on the input,
    including 3x3, 7x7x3 convolutions, and a max-pooling layer, before
    concatenating the results along the channel dimension.

    Examples
    --------
    >>> import torch
    >>> from deepchem.models.torch_models.inceptionv3 import InceptionD
    >>> layer = InceptionD(768)
    >>> x = torch.randn(5, 768, 17, 17)
    >>> output = layer(x)
    >>> output.shape
    torch.Size([5, 1280, 8, 8])
    """

    def __init__(
        self,
        in_channels: int,
    ) -> None:
        super(InceptionD, self).__init__()
        self.branch3x3_1 = BasicConv2d(in_channels,
                                       192,
                                       kernel_size=(1, 1),
                                       stride=(1, 1),
                                       padding=(0, 0))
        self.branch3x3_2 = BasicConv2d(192,
                                       320,
                                       kernel_size=(3, 3),
                                       stride=(2, 2),
                                       padding=(0, 0))

        self.branch7x7x3_1 = BasicConv2d(in_channels,
                                         192,
                                         kernel_size=(1, 1),
                                         stride=(1, 1),
                                         padding=(0, 0))
        self.branch7x7x3_2 = BasicConv2d(192,
                                         192,
                                         kernel_size=(1, 7),
                                         stride=(1, 1),
                                         padding=(0, 3))
        self.branch7x7x3_3 = BasicConv2d(192,
                                         192,
                                         kernel_size=(7, 1),
                                         stride=(1, 1),
                                         padding=(3, 0))
        self.branch7x7x3_4 = BasicConv2d(192,
                                         192,
                                         kernel_size=(3, 3),
                                         stride=(2, 2),
                                         padding=(0, 0))

        # self.maxpool = nn.MaxPool2d((3, 3), (2, 2))

    def forward(self, x):
        """
        Forward pass for the InceptionD module.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape (N, in_channels, H, W).

        Returns
        -------
        torch.Tensor
            Output tensor of shape (N, out_channels, H, W),
            where out_channels is the total number of output channels
            from all branches combined.
        """
        branch3x3 = self.branch3x3_1(x)
        branch3x3 = self.branch3x3_2(branch3x3)

        branch7x7x3 = self.branch7x7x3_1(x)
        branch7x7x3 = self.branch7x7x3_2(branch7x7x3)
        branch7x7x3 = self.branch7x7x3_3(branch7x7x3)
        branch7x7x3 = self.branch7x7x3_4(branch7x7x3)

        branch_pool = F.max_pool2d(x, kernel_size=3, stride=2)

        out = [branch3x3, branch7x7x3, branch_pool]
        output = torch.cat(out, 1)

        return output


class InceptionE(nn.Module):
    """
    InceptionE module as part of the InceptionV3 architecture.
    This module performs parallel operations on the input,
    including 1x1, 3x3, and double 3x3 convolutions, along with an
    average pooling branch, then concatenates the results along the
    channel dimension.

    Examples
    --------
    >>> import torch
    >>> from deepchem.models.torch_models.inceptionv3 import InceptionE
    >>> layer = InceptionE(1280)
    >>> x = torch.randn(5,1280, 8, 8)
    >>> output = layer(x)
    >>> output.shape
    torch.Size([5, 2048, 8, 8])
    """

    def __init__(
        self,
        in_channels: int,
    ) -> None:
        super(InceptionE, self).__init__()
        self.branch1x1 = BasicConv2d(in_channels,
                                     320,
                                     kernel_size=(1, 1),
                                     stride=(1, 1),
                                     padding=(0, 0))

        self.branch3x3_1 = BasicConv2d(in_channels,
                                       384,
                                       kernel_size=(1, 1),
                                       stride=(1, 1),
                                       padding=(0, 0))
        self.branch3x3_2a = BasicConv2d(384,
                                        384,
                                        kernel_size=(1, 3),
                                        stride=(1, 1),
                                        padding=(0, 1))
        self.branch3x3_2b = BasicConv2d(384,
                                        384,
                                        kernel_size=(3, 1),
                                        stride=(1, 1),
                                        padding=(1, 0))

        self.branch3x3dbl_1 = BasicConv2d(in_channels,
                                          448,
                                          kernel_size=(1, 1),
                                          stride=(1, 1),
                                          padding=(0, 0))
        self.branch3x3dbl_2 = BasicConv2d(448,
                                          384,
                                          kernel_size=(3, 3),
                                          stride=(1, 1),
                                          padding=(1, 1))
        self.branch3x3dbl_3a = BasicConv2d(384,
                                           384,
                                           kernel_size=(1, 3),
                                           stride=(1, 1),
                                           padding=(0, 1))
        self.branch3x3dbl_3b = BasicConv2d(384,
                                           384,
                                           kernel_size=(3, 1),
                                           stride=(1, 1),
                                           padding=(1, 0))

        # self.avgpool = nn.AvgPool2d((3, 3), (1, 1), (1, 1))
        self.branch_pool = BasicConv2d(in_channels, 192, kernel_size=1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass for the InceptionE module.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape (N, in_channels, H, W).

        Returns
        -------
        torch.Tensor
            Output tensor of shape (N, out_channels, H, W),
            where out_channels is the total number of output channels
            from all branches combined.
        """
        branch1x1 = self.branch1x1(x)

        branch3x3 = self.branch3x3_1(x)
        branch3x3 = torch.cat(
            [self.branch3x3_2a(branch3x3),
             self.branch3x3_2b(branch3x3)], 1)

        branch3x3dbl = self.branch3x3dbl_1(x)
        branch3x3dbl = self.branch3x3dbl_2(branch3x3dbl)
        branch3x3dbl = torch.cat([
            self.branch3x3dbl_3a(branch3x3dbl),
            self.branch3x3dbl_3b(branch3x3dbl)
        ], 1)

        branch_pool = F.avg_pool2d(x, kernel_size=3, stride=1, padding=1)
        branch_pool = self.branch_pool(branch_pool)

        out = [branch1x1, branch3x3, branch3x3dbl, branch_pool]
        output = torch.cat(out, 1)

        return output


class InceptionAux(nn.Module):
    """
    An auxiliary classifier module used in the InceptionV3 architecture.
    This module is intended to provide an auxiliary output for
    intermediate supervision during training.

    Examples
    --------
    >>> import torch
    >>> from deepchem.models.torch_models.inceptionv3 import InceptionAux
    >>> layer = InceptionAux(768, num_classes=3)
    >>> x = torch.randn(5, 768, 17, 17)
    >>> output = layer(x)
    >>> output.shape
    torch.Size([5, 3])
    """

    def __init__(
        self,
        in_channels: int,
        num_classes: int,
    ) -> None:
        super().__init__()
        # self.avgpool1 = nn.AvgPool2d((5, 5), (3, 3))
        self.conv0 = BasicConv2d(in_channels,
                                 128,
                                 kernel_size=(1, 1),
                                 stride=(1, 1),
                                 padding=(0, 0))
        self.conv1 = BasicConv2d(128,
                                 768,
                                 kernel_size=(5, 5),
                                 stride=(1, 1),
                                 padding=(0, 0))
        self.conv1.stddev = 0.01  # type: ignore[assignment]
        # self.avgpool2 = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(768, num_classes)
        self.fc.stddev = 0.001  # type: ignore[assignment]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass for the InceptionAux module.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape (N, in_channels, H, W).

        Returns
        -------
        torch.Tensor
            Output tensor of shape (N, num_classes) representing class
            scores for each input.
        """
        # N x 768 x 17 x 17
        x = F.avg_pool2d(x, kernel_size=5, stride=3)
        # N x 768 x 5 x 5
        x = self.conv0(x)
        # N x 128 x 5 x 5
        x = self.conv1(x)
        # N x 768 x 1 x 1
        x = F.adaptive_avg_pool2d(x, (1, 1))
        # N x 768 x 1 x 1
        x = torch.flatten(x, 1)
        # N x 768
        x = self.fc(x)
        # N x 3 (num_classes)

        return x


class InceptionV3Model(TorchModel):
    """
    Implementation of the InceptionV3 model architecture for image
    classification, modified for use with the DeepVariant framework
    in DeepChem.

    It builds on the original Inception design by utilizing a
    network-in-network approach, where convolutional filters of various
    sizes (e.g., 1x1, 3x3, 5x5) are applied in parallel within each module.
    This enables the model to capture features at multiple scales.
    InceptionV3 has factorized convolutions (breaking down larger
    convolutions into smaller ones, like 3x3 into 1x3 and 3x1)
    and the use of auxiliary classifiers that assist the model’s training
    by acting as regularizers. It uses dimensionality reduction to control
    the computational complexity.This model supports custom learning rate
    schedules with warmup and decay steps, utilizing the RMSProp optimizer.

    Examples
    --------
    >>> from deepchem.models.torch_models import InceptionV3Model
    >>> import deepchem as dc
    >>> import numpy as np
    >>> model = InceptionV3Model()
    >>> input_shape = (5, 6, 299, 299)
    >>> input_samples = np.random.randn(*input_shape).astype(np.float32)
    >>> output_samples = np.random.randint(0, 3, (5,)).astype(np.int64)
    >>> one_hot_output_samples = np.eye(3)[output_samples]
    >>> dataset = dc.data.ImageDataset(input_samples, one_hot_output_samples)
    >>> loss = model.fit(dataset, nb_epoch=1)
    >>> predictions = model.predict(dataset)
    >>> predictions.shape
    (5, 3)

    """

    def __init__(self,
                 in_channels=6,
                 warmup_steps=10000,
                 learning_rate=0.064,
                 dropout_rate=0.2,
                 decay_rate=0.94,
                 decay_steps=2,
                 rho=0.9,
                 momentum=0.9,
                 epsilon=1.0,
                 **kwargs):
        # weight_decay = 0.00004

        # Initialize the InceptionV3 model architecture
        model = InceptionV3(num_classes=3,
                            in_channels=in_channels,
                            aux_logits=False,
                            dropout_rate=dropout_rate)

        loss = CategoricalCrossEntropy()

        # Define optimizer as DeepChem's RMSProp
        optimizer = RMSProp(
            learning_rate=learning_rate,
            momentum=momentum,
            decay=rho,  # Using decay as rho
            epsilon=epsilon)

        # Custom attributes for learning rate decay and warmup
        self.learning_rate = learning_rate
        self.decay_rate = decay_rate
        self.decay_steps = decay_steps
        self.warmup_steps = warmup_steps
        self.current_step = 0

        # Initialize base TorchModel
        super(InceptionV3Model, self).__init__(model=model,
                                               optimizer=optimizer,
                                               loss=loss,
                                               **kwargs)

    def adjust_learning_rate(self):
        """
        Adjusts learning rate manually based on warmup and decay steps.
        """
        if self.current_step < self.warmup_steps:
            lr = self.learning_rate * (self.current_step / self.warmup_steps)
        else:
            decay_factor = self.decay_rate**(self.current_step //
                                             self.decay_steps)
            lr = self.learning_rate * decay_factor

        self.optimizer.learning_rate = lr

    def fit(self,
            dataset: Dataset,
            nb_epoch: int = 10,
            max_checkpoints_to_keep: int = 5,
            checkpoint_interval: int = 1000,
            deterministic: bool = False,
            restore: bool = False,
            variables: Optional[List[torch.nn.Parameter]] = None,
            loss: Optional[LossFn] = None,
            callbacks: Union[Callable, List[Callable]] = [],
            all_losses: Optional[List[float]] = None) -> float:
        """
        Trains the model on the given dataset, adjusting learning rate
        with warmup and decay.

        Parameters
        ----------
        dataset: Dataset
            Dataset to be used for training.
        nb_epoch: int, optional (default 1)
            Number of epochs to train the model.
        max_checkpoints_to_keep: int, optional
            Number of checkpoints to keep.
        checkpoint_interval: int, optional
            Interval for saving checkpoints.
        deterministic: bool, optional
            If True, runs in deterministic mode.
        restore: bool, optional
            If True, restores the model from the last checkpoint.
        variables: list, optional
            List of parameters to train.
        loss: callable, optional
            Custom loss function.
        callbacks: callable or list of callables, optional
            Callbacks to run during training.
        all_losses: list of floats, optional
            List to store all losses during training.

        Returns
        -------
        float
            The final loss value after training.
        """
        if all_losses is None:
            all_losses = []

        for epoch in range(nb_epoch):

            self.current_step = epoch
            self.adjust_learning_rate(
            )  # Adjust learning rate before each epoch

            epoch_loss = super(InceptionV3Model, self).fit(
                dataset,
                nb_epoch=1,
                max_checkpoints_to_keep=max_checkpoints_to_keep,
                checkpoint_interval=checkpoint_interval,
                deterministic=deterministic,
                restore=restore,
                variables=variables,
                loss=loss,
                callbacks=callbacks,
                all_losses=all_losses)
            all_losses.append(epoch_loss)

        return all_losses[-1] if all_losses else 0.0

    def save(self):
        """Saves model to disk using joblib."""
        save_to_disk(self.model, self.get_model_filename(self.model_dir))

    def reload(self):
        """Loads model from joblib file on disk."""
        self.model = load_from_disk(self.get_model_filename(self.model_dir))



from deepchem.data import Dataset


class VocabularyBuilder():
    """Abstract class for building a vocabulary from a dataset."""

    def build(self, dataset: Dataset):
        """Builds vocabulary from a dataset

        Parameters
        ----------
        dataset: Dataset
            dataset to build vocabulary from.
        """
        raise NotImplementedError()

    @classmethod
    def load(cls, fname: str):
        """Loads vocabulary from the specified file

        Parameters
        ----------
        fname: str
            Path containing pre-build vocabulary.
        """
        raise NotImplementedError()

    def save(self, fname: str):
        """Dump vocabulary to the specified file.

        Parameters
        ----------
        fname: str
            A json file fname to save vocabulary.
        """
        raise NotImplementedError()

    def extend(self, dataset: Dataset):
        """Extends vocabulary from a dataset

        Parameters
        ----------
        dataset: Dataset
            dataset used for extending vocabulary
        """
        raise NotImplementedError()



"""Part of this code adopted from https://github.com/diffqc/dqc"""
from __future__ import annotations
from abc import abstractmethod
from typing import Optional, Dict, Any, List, Union, Tuple
import torch
from deepchem.utils.differentiation_utils import EditableModule, minimize, equilibrium, set_default_option
from deepchem.utils.dft_utils import BaseSystem, BaseQCCalc, SpinParam
from deepchem.utils.dft_utils.config import config


class SCF_QCCalc(BaseQCCalc):
    """
    Performing Restricted or Unrestricted self-consistent field
    iteration (e.g. Hartree-Fock or Density Functional Theory)

    Examples
    --------
    >>> from deepchem.utils.dft_utils import SCF_QCCalc, BaseSCFEngine
    >>> from deepchem.utils.dft_utils import SpinParam

    # Define the engine
    >>> class engine(BaseSCFEngine):
    ...     def polarised():
    ...         return False
    ...     def dm2energy(self, dm: torch.Tensor | SpinParam[torch.Tensor]) -> torch.Tensor:
    ...        return dm * 1.1
    >>> myEngine = engine()
    >>> a = SCF_QCCalc(myEngine)
    >>> a.dm2energy(torch.tensor([1.1]))
    tensor([1.2100])

    """

    def __init__(self, engine: BaseSCFEngine, variational: bool = False):
        """Initialize the SCF calculation.

        Parameters
        ----------
        engine: BaseSCFEngine
            SCF engine
        variational: bool
            If True, then use optimization of the free orbital
            parameters to find the minimum energy. Otherwise, use
            self-consistent iterations.

        """
        self._engine = engine
        self._polarized = engine.polarized
        self._shape = self._engine.shape
        self.dtype = self._engine.dtype
        self.device = self._engine.device
        self._has_run = False
        self._variational = variational

    def get_system(self) -> BaseSystem:
        """Returns the system in the QC calculation

        Returns
        -------
        BaseSystem
            System that is being calculated.

        """
        return self._engine.get_system()

    def run(  # type: ignore[override]
            self,
            dm0: Optional[
                Union[str, torch.Tensor,
                      SpinParam[torch.Tensor]]] = "1e",  # type: ignore
            eigen_options: Optional[Dict[str, Any]] = None,
            fwd_options: Optional[Dict[str, Any]] = None,
            bck_options: Optional[Dict[str, Any]] = None) -> BaseQCCalc:
        """Run the calculation.

        Note that this method can be invoked several times for one
        object to try for various self-consistent options to reach
        convergence.

        Parameters
        ----------
        dm0: Optional[Union[str, torch.Tensor, SpinParam[torch.Tensor]]]
            Initial density matrix. If it is a string, it can be
            "1e" to use the 1-electron Hamiltonian to generate the
            initial density matrix.
        eigen_options: Optional[Dict[str, Any]]
            Options for the diagonalization (i.e. eigendecomposition).
        fwd_options: Optional[Dict[str, Any]]
            Options for the forward iteration (i.e. the iteration
            to find the minimum energy).
        bck_options: Optional[Dict[str, Any]]
            Options for the backward iteration (i.e. the iteration
            to find the minimum energy).

        """
        # get default options
        if not self._variational:
            fwd_defopt = {
                "method": "broyden1",
                "alpha": -0.5,
                "maxiter": 50,
                "verbose": config.VERBOSE > 0,
            }
        else:
            fwd_defopt = {
                "method": "gd",
                "step": 1e-2,
                "maxiter": 5000,
                "f_rtol": 1e-10,
                "x_rtol": 1e-10,
                "verbose": config.VERBOSE > 0,
            }
        bck_defopt = {
            "posdef": True,
        }

        # setup the default options
        if eigen_options is None:
            eigen_options = {"method": "exacteig"}
        if fwd_options is None:
            fwd_options = {}
        if bck_options is None:
            bck_options = {}
        fwd_options = set_default_option(fwd_defopt, fwd_options)
        bck_options = set_default_option(bck_defopt, bck_options)

        # save the eigen_options for use in diagonalization
        self._engine.set_eigen_options(eigen_options)

        # set up the initial self-consistent param guess
        if dm0 is None:
            dm = self._get_zero_dm()
        elif isinstance(dm0, str):
            if dm0 == "1e":  # initial density based on 1-electron Hamiltonian
                dm = self._get_zero_dm()
                scp0 = self._engine.dm2scp(dm)
                dm = self._engine.scp2dm(scp0)
            else:
                raise RuntimeError("Unknown dm0: %s" % dm0)
        else:
            dm = SpinParam.apply_fcn(lambda dm0: dm0.detach(), dm0)

        # making it spin param for polarized and tensor for nonpolarized
        if isinstance(dm, torch.Tensor) and self._polarized:
            dm_u = dm * 0.5
            dm_d = dm * 0.5
            dm = SpinParam(u=dm_u, d=dm_d)
        elif isinstance(dm, SpinParam) and not self._polarized:
            dm = dm.u + dm.d

        if not self._variational:
            scp0 = self._engine.dm2scp(dm)

            # do the self-consistent iteration
            scp = equilibrium(fcn=self._engine.scp2scp,
                              y0=scp0,
                              bck_options={**bck_options},
                              **fwd_options)

            # post-process parameters
            self._dm = self._engine.scp2dm(scp)
        else:
            system = self.get_system()
            h = system.get_hamiltonian()
            orb_weights = system.get_orbweight(polarized=self._polarized)
            norb = SpinParam.apply_fcn(lambda orb_weights: len(orb_weights),
                                       orb_weights)

            def dm2params(dm: Union[torch.Tensor, SpinParam[torch.Tensor]]) -> \
                    Tuple[torch.Tensor, torch.Tensor]:
                """Convert the density matrix to the atomic orbital parameters.

                Parameters
                ----------
                dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
                    Density matrix. It is tensor if restricted, and SpinParam
                    of tensor if unrestricted.

                Returns
                -------
                Tuple[torch.Tensor, torch.Tensor]
                    Atomic orbital parameters and coefficients.

                """
                pc = SpinParam.apply_fcn(
                    lambda dm, norb: h.dm2ao_orb_params(SpinParam.sum(dm),
                                                        norb=norb), dm, norb)
                p = SpinParam.apply_fcn(lambda pc: pc[0], pc)
                c = SpinParam.apply_fcn(lambda pc: pc[1], pc)
                params = self._engine.pack_aoparams(p)
                coeffs = self._engine.pack_aoparams(c)
                return params, coeffs

            def params2dm(params: torch.Tensor, coeffs: torch.Tensor) \
                    -> Union[torch.Tensor, SpinParam[torch.Tensor]]:
                """Convert the atomic orbital parameters to the density matrix.

                Parameters
                ----------
                params: torch.Tensor
                    Atomic orbital parameters.
                coeffs: torch.Tensor
                    Atomic orbital coefficients.

                Returns
                -------
                Union[torch.Tensor, SpinParam[torch.Tensor]]
                    Density matrix. It is tensor if restricted, and SpinParam
                    of tensor if unrestricted.

                """
                p: Union[
                    torch.Tensor,
                    SpinParam[torch.Tensor]] = self._engine.unpack_aoparams(
                        params)
                c: Union[
                    torch.Tensor,
                    SpinParam[torch.Tensor]] = self._engine.unpack_aoparams(
                        coeffs)

                dm = SpinParam.apply_fcn(
                    lambda p, c, orb_weights: h.ao_orb_params2dm(
                        p, c, orb_weights, with_penalty=None), p, c,
                    orb_weights)
                return dm

            params0, coeffs0 = dm2params(dm)
            params0 = params0.detach()
            coeffs0 = coeffs0.detach()
            min_params0: torch.Tensor = minimize(
                fcn=self._engine.aoparams2ene,
                # random noise to add the chance of it gets to the minimum, not
                # a saddle point
                y0=params0 + torch.randn_like(params0) * 0.03 / params0.numel(),
                params=(
                    coeffs0,
                    None,
                ),  # coeffs & with_penalty
                bck_options={
                    **bck_options
                },
                **fwd_options).detach()

            if torch.is_grad_enabled():
                # If the gradient is required, then put it through the minimization
                # one more time with penalty on the parameters.
                # penalty is to keep the Hamiltonian invertible, stabilizing
                # inverse.
                # Without the penalty, the Hamiltonian could have 0 eigenvalues
                # because of the overparameterization of the aoparams.
                min_dm = params2dm(min_params0, coeffs0)
                params0, coeffs0 = dm2params(min_dm)
                min_params0 = minimize(
                    fcn=self._engine.aoparams2ene,
                    y0=params0,
                    params=(
                        coeffs0,
                        1e-1,
                    ),  # coeffs & with_penalty
                    bck_options={**bck_options},
                    method="gd",
                    step=0,
                    maxiter=0)

            self._dm = params2dm(min_params0, coeffs0)

        self._has_run = True
        return self

    def energy(self) -> torch.Tensor:
        """Obtain the energy of the system.

        Returns
        -------
        torch.Tensor
            Energy of the system.

        """
        assert self._has_run
        return self._engine.dm2energy(self._dm)

    def aodm(self) -> Union[torch.Tensor, SpinParam[torch.Tensor]]:
        """
        Returns the density matrix in atomic orbital. For polarized
        case, it returns a SpinParam of 2 tensors representing the
        density matrices for spin-up and spin-down.

        Returns
        -------
        Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix in atomic orbital. Shape: (nao, nao)

        """
        assert self._has_run
        return self._dm

    def dm2energy(self, dm: Union[torch.Tensor, SpinParam[torch.Tensor]]):
        """Calculate the energy from the given density matrix.

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Input density matrix. It is tensor if restricted, and
            SpinParam of tensor if unrestricted.

        Returns
        -------
        torch.Tensor
            Tensor that represents the energy given the energy.

        """
        assert (isinstance(dm, torch.Tensor) and not self._polarized) or \
            (isinstance(dm, SpinParam) and self._polarized), type(dm)
        return self._engine.dm2energy(dm)

    def _get_zero_dm(self) -> Union[SpinParam[torch.Tensor], torch.Tensor]:
        """get the initial dm that are all zeros

        Returns
        -------
        Union[SpinParam[torch.Tensor], torch.Tensor]
            Initial density matrix. It is tensor if restricted,
            and SpinParam of tensor if unrestricted.

        """
        if not self._polarized:
            return torch.zeros(self._shape,
                               dtype=self.dtype,
                               device=self.device)
        else:
            dm0_u = torch.zeros(self._shape,
                                dtype=self.dtype,
                                device=self.device)
            dm0_d = torch.zeros(self._shape,
                                dtype=self.dtype,
                                device=self.device)
            return SpinParam(u=dm0_u, d=dm0_d)


class BaseSCFEngine(EditableModule):
    """Base class for the SCF engine.

    This class is the base class for the self-consistent field (SCF) engine.
    It is used to perform the SCF iteration to find the minimum energy of the
    system.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import SCF_QCCalc, BaseSCFEngine
    >>> from deepchem.utils.dft_utils import SpinParam

    # Define the engine
    >>> class engine(BaseSCFEngine):
    ...     def polarised():
    ...         return False
    ...     def dm2energy(self, dm: torch.Tensor | SpinParam[torch.Tensor]) -> torch.Tensor:
    ...        return dm * 1.1
    >>> myEngine = engine()
    >>> a = SCF_QCCalc(myEngine)
    >>> a.dm2energy(torch.tensor([1.1]))
    tensor([1.2100])

    """

    @property
    @abstractmethod
    def polarized(self) -> bool:
        """If the system is polarized or not.

        Returns
        -------
        bool
            True if the system is polarized, False otherwise.

        """
        pass

    @property
    @abstractmethod
    def shape(self):
        """Shape of the density matrix in this engine.

        Returns
        -------
        Any
            shape of the density matrix in this engine.

        """
        pass

    @property
    @abstractmethod
    def dtype(self) -> torch.dtype:
        """dtype of the tensors in this engine.

        Returns
        -------
        torch.dtype
            dtype of the tensors in this engine.

        """
        pass

    @property
    @abstractmethod
    def device(self) -> torch.device:
        """Device of the tensors in this engine.

        Returns
        -------
        torch.device
            device of the tensors in this engine.

        """
        pass

    @abstractmethod
    def get_system(self) -> BaseSystem:
        """System involved in the engine.

        Returns
        -------
        BaseSystem
            system involved in the engine.

        """
        pass

    @abstractmethod
    def dm2energy(
            self, dm: Union[torch.Tensor,
                            SpinParam[torch.Tensor]]) -> torch.Tensor:
        """Calculate the energy from the given density matrix.

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            input density matrix. It is tensor if restricted, and
            SpinParam of tensor if unrestricted.

        Returns
        -------
        torch.Tensor
            Energy of the given density matrix.

        """
        pass

    @abstractmethod
    def dm2scp(
            self, dm: Union[torch.Tensor,
                            SpinParam[torch.Tensor]]) -> torch.Tensor:
        """
        Convert the density matrix into the self-consistent parameter (scp).
        Self-consistent parameter is defined as the parameter that is put into
        the equilibrium function, i.e. y in `y = f(y, x)`.

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            input density matrix. It is tensor if restricted, and
            SpinParam of tensor if unrestricted.

        Returns
        -------
        torch.Tensor
            Self-consistent parameter.

        """
        pass

    @abstractmethod
    def scp2dm(
            self,
            scp: torch.Tensor) -> Union[torch.Tensor, SpinParam[torch.Tensor]]:
        """Calculate the density matrix from the given self-consistent parameter (scp).

        Parameters
        ----------
        scp: torch.Tensor
            Input self-consistent parameter.

        Returns
        -------
        Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix. It is tensor if restricted, and SpinParam of
            tensor if unrestricted.

        """
        pass

    @abstractmethod
    def scp2scp(self, scp: torch.Tensor) -> torch.Tensor:
        """Calculate the next self-consistent parameter (scp) for the
        next iteration from the previous scp.

        Parameters
        ----------
        scp: torch.Tensor
            Previous self-consistent parameter.

        Returns
        -------
        torch.Tensor
            Next self-consistent parameter.

        """
        pass

    @abstractmethod
    def aoparams2ene(self,
                     aoparams: torch.Tensor,
                     aocoeffs: torch.Tensor,
                     with_penalty: Optional[float] = None) -> torch.Tensor:
        """Calculate the energy from the given atomic orbital parameters
        and coefficients.

        Parameters
        ----------
        aoparams: torch.Tensor
            Atomic orbital parameters.
        aocoeffs: torch.Tensor
            Atomic orbital coefficients.
        with_penalty: Optional[float]
            Penalty factor.

        Returns
        -------
        torch.Tensor
            Energy from the given atomic orbital parameters and
            coefficients.

        """
        pass

    @abstractmethod
    def aoparams2dm(self, aoparams: torch.Tensor, aocoeffs: torch.Tensor,
                    with_penalty: Optional[float] = None) -> \
            Tuple[Union[torch.Tensor, SpinParam[torch.Tensor]], Optional[torch.Tensor]]:
        """Calculate the density matrix and the penalty from the given
        atomic orbital parameters and coefficients.

        Parameters
        ----------
        aoparams: torch.Tensor
            Atomic orbital parameters.
        aocoeffs: torch.Tensor
            Atomic orbital coefficients.
        with_penalty: Optional[float]
            Penalty factor.

        Returns
        -------
        Tuple[Union[torch.Tensor, SpinParam[torch.Tensor]], Optional[torch.Tensor]]
            Density matrix and the penalty from the given atomic
            orbital parameters and coefficients.

        """
        pass

    @abstractmethod
    def pack_aoparams(
            self, aoparams: Union[torch.Tensor,
                                  SpinParam[torch.Tensor]]) -> torch.Tensor:
        """Pack the ao params into a single tensor.

        Parameters
        ----------
        aoparams: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Input atomic orbital parameters. It is tensor if
            restricted, and SpinParam of tensor if unrestricted.

        Returns
        -------
        torch.Tensor
            Packed atomic orbital parameters.

        """
        pass

    @abstractmethod
    def unpack_aoparams(
            self, aoparams: torch.Tensor
    ) -> Union[torch.Tensor, SpinParam[torch.Tensor]]:
        """Unpack the ao params into a tensor or SpinParam of tensor.

        Parameters
        ----------
        aoparams: torch.Tensor
            Input packed atomic orbital parameters.

        Returns
        -------
        Union[torch.Tensor, SpinParam[torch.Tensor]]
            Atomic orbital parameters. It is tensor if restricted,
            and SpinParam of tensor if unrestricted.

        """
        pass

    @abstractmethod
    def set_eigen_options(self, eigen_options: Dict[str, Any]) -> None:
        """Set the options for the diagonalization (i.e. eigendecomposition).

        Parameters
        ----------
        eigen_options: Dict[str, Any]
            Options for the diagonalization.

        """
        pass

    @abstractmethod
    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """List all the names of parameters used in the given method.

        Parameters
        ----------
        methodname: str
            Name of the method.
        prefix: str
            Prefix to be added to the parameter names.

        Returns
        -------
        List[str]
            List of parameter names.

        """
        pass
